import logging
from collections import OrderedDict
import torch
import torch.nn as nn
from . import lr_scheduler
from . import networks
from .base_model import BaseModel
from apex import amp
import apex

logger = logging.getLogger('base')


class SRModel(BaseModel):
    def __init__(self, opt):
        super(SRModel, self).__init__(opt)
        train_opt = opt['train']
        if self.is_train:
            self.netG.train()

            # set-up loss: pixel loss
            loss_type = train_opt['pixel_criterion']
            if loss_type == 'l1':
                print('l1 loss: l1')
                self.cri_pix = nn.L1Loss().to(self.device)
            elif loss_type == 'l2':
                print('l2: MSE loss')
                self.cri_pix = nn.MSELoss().to(self.device)
            else:
                raise NotImplementedError('Loss type [{:s}] is not recognized.'.format(loss_type))
            # weights to be multiplied to pixel loss
            self.l_pix_w = train_opt['pixel_weight']

            # set-up optimizers
            wd_G = train_opt['weight_decay_G'] if train_opt['weight_decay_G'] else 0
            optim_params = []
            for k, v in self.netG.named_parameters():  # can optimize for a part of the model
                if v.requires_grad:
                    optim_params.append(v)
                else:
                    logger.warning('Params [{:s}] will not optimize.'.format(k))
            if opt["network_G"]["which_model_G"] == "fsrcnn":
                self.optimizer_G = torch.optim.Adam([
                        {'params': self.netG.first_part.parameters()},
                        {'params': self.netG.mid_part.parameters()},
                        {'params': self.netG.last_part.parameters(), 'lr': train_opt['lr_G'] * 0.1}
                        ], lr=train_opt['lr_G'])
            else:
                self.optimizer_G = torch.optim.Adam(
                    optim_params, lr=train_opt['lr_G'], weight_decay=wd_G, betas=(train_opt['beta1_G'], train_opt['beta2_G']))
            self.optimizers.append(self.optimizer_G)

            # set-up schedulers
            if train_opt['lr_scheme'] == 'MultiStepLR':
                for optimizer in self.optimizers:
                    self.schedulers.append(
                        lr_scheduler.MultiStepLR_Restart(optimizer, train_opt['lr_steps'],
                                                         restarts=train_opt['restarts'],
                                                         weights=train_opt['restart_weights'],
                                                         gamma=train_opt['lr_gamma'],
                                                         clear_state=train_opt['clear_state']))
            elif train_opt['lr_scheme'] == 'CosineAnnealingLR':
                for optimizer in self.optimizers:
                    self.schedulers.append(
                        lr_scheduler.CosineAnnealingLR_Restart(optimizer, train_opt['T_period'],
                                                        eta_min=train_opt['eta_min'],
                                                        restarts=train_opt['restarts'],
                                                        weights=train_opt['restart_weights']))
            else:
                raise NotImplementedError('Choose MultiStepLR or CosineAnnealingLR')
            self.log_dict = OrderedDict()
        self.print_network()
        self.load()

    def initialize_amp(self):
        self.netG, self.optimizer_G = amp.initialize(self.netG, self.optimizer_G, opt_level=self.opt['opt_level'])
        if self.opt['gpu_ids']:
            assert torch.cuda.is_available()
            self.netG = nn.DataParallel(self.netG)
    
    def run_test(self, data): 
        need_HR = False if self.opt['datasets']['dataroot_HR'] is None else True
        has_mask = False if self.opt['datasets']['maskroot_HR'] is None else True
        self.feed_test_data(data, need_HR=need_HR)
        self.test(data)  # test
        visuals = self.get_current_visuals(data, maskOn=has_mask, need_HR=need_HR)
        data['volume'] = visuals['SR'][None,None]
        return data 

    def test(self, data):

        self.netG_eval.eval()
        if self.opt['precision'] == 'fp16':
            var_L_eval = self.var_L.half()
        else:
            var_L_eval = self.var_L

        opt_val = self.opt['datasets']['val']
        # HR slice num
        num_HR = int (data['LR'].size(2) * self.opt['scale'])
        # HR slice overlap
        HR_ot = int(self.opt['scale'] * self.ot)

        if not opt_val['need_voxels']:
            print('---- Full size image prediction -----')
            pt, H, W = int(self.var_L.size(2)*self.opt['scale']), self.var_L.size(3), self.var_L.size(4)
            self.fake_H = torch.empty(1, 1,  num_HR, H, W, device=self.device)
            if self.opt['precision'] == 'fp16':
                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, H, W, dtype=torch.half, device=self.device)
            else:
                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, H, W, device=self.device)
            
            # mask, record 1 when there is value in this pixel
            stitch_mask = torch.zeros_like(self.fake_H, device=self.device)
            with torch.no_grad():
                if opt_val['full_volume']:
                    self.fake_H = self.netG_eval(var_L_eval)
                else:
                    for i in range(0, self.nt):
                        fake_H_in_chunks[[i],...] = self.netG_eval(var_L_eval[[i],...])
                    # stitch volume in z-direction: n-1 volumes
                    for i in range(0, self.nt - 1):
                        ts, te = i * (pt - HR_ot), i * (pt - HR_ot) + pt
                        # `stitch_mask` and `fake_H` are fp32 for better numerical stability - fp16 inference results are casted into fp32
                        self.fake_H[0, 0, ts:te, :, :] = \
                        (self.fake_H[0, 0, ts:te, :, :] * stitch_mask[0, 0, ts:te, :, :] +
                        fake_H_in_chunks[i,...].float() * (2 - stitch_mask[0, 0, ts:te, :, :])) / 2
                        stitch_mask[0, 0, ts:te, :, :] = 1.
                    # the last volume
                    self.fake_H[0, 0, -pt:, :, :] = \
                        (self.fake_H[0, 0, -pt:, :, :] * stitch_mask[0, 0, -pt:, :, :] +
                        fake_H_in_chunks[-1,...].float() * (2 - stitch_mask[0, 0, -pt:, :, :])) / 2
        else:
            if opt_val['need_voxels'] and not opt_val['need_voxels']['tile_x_y']:
                # print('---- Random voxel prediction-----')
                self.fake_H = self.netG_eval(var_L_eval).float()
            elif opt_val['need_voxels'] and opt_val['need_voxels']['tile_x_y']:
                # print('---- Sliding tile prediction ----')
                # print('case has to be tiled back:', self.var_L.shape)
                pt, H, W = opt_val['slice_size'], data['LR'].size(3), data['LR'].size(4)
                pt = int(pt * self.opt['scale']) # 32*1 = 32
                self.fake_H = torch.empty(1, 1,  num_HR, H, W, device=self.device)
                # get predictions
                with torch.no_grad():
                    for row in range(0, data['LR'].size(3), opt_val['need_voxels']['tile_size']):
                        for column in range(0, data['LR'].size(3), opt_val['need_voxels']['tile_size']):
                            # [1, 1, 330, 64, 64]
                            LR_chunked = var_L_eval[:, :, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']]
                            if self.opt['precision'] == 'fp16':
                                # [12, 1, 32, 64, 64]
                                tmp_chunk_along_z = torch.empty(self.nt, 1, pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'],
                                                    dtype=torch.half, device=self.device)
                            else:
                                tmp_chunk_along_z = torch.empty(self.nt, 1, pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'],
                                                    device=self.device)

                            # iterate over number of blocks to get predictions
                            for i in range(0, self.nt - 1):
                                tmp_chunk_along_z[i, :, :, :, :] = self.netG_eval(LR_chunked[:, :, i*(pt-self.ot):i*(pt-self.ot)+pt, :, :])
                            tmp_chunk_along_z[-1, :, :, :, :] = self.netG_eval(LR_chunked[:, :, -pt:, :, :])

                            # reconstruct the volume along z [1, 1, 330, 64, 64]
                            reconstructed_z = torch.empty(1, 1, num_HR, opt_val['need_voxels']['tile_size'],
                                                        opt_val['need_voxels']['tile_size'], device=self.device)
                            # stitch volume along z direction
                            stitch_mask = torch.zeros_like(reconstructed_z, device=self.device)
                            for i in range(0, self.nt - 1):
                                ts, te = i * (pt - HR_ot), i * (pt - HR_ot) + pt
                                reconstructed_z[0, 0, ts:te, :, :] = (reconstructed_z[0, 0, ts:te, :, :] * stitch_mask[0, 0, ts:te, :, :] + 
                                                                        tmp_chunk_along_z[i,...].float() * (2 - stitch_mask[0, 0, ts:te, :, :])) / 2
                                stitch_mask[0, 0, ts:te, :, :] = 1.
                            # stich last volume
                            reconstructed_z[0, 0, -pt:, :, :] = \
                                (reconstructed_z[0, 0, -pt:, :, :] * stitch_mask[0, 0, -pt:, :, :] +
                                tmp_chunk_along_z[-1,...].float() * (2 - stitch_mask[0, 0, -pt:, :, :])) / 2
                            # accumulate volume together
                            self.fake_H[0, 0, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']] = reconstructed_z
            else:
                raise ValueError('Unknown tiling case found in SR model!')
        self.netG.train()

    def optimize_parameters(self, step):
        self.optimizer_G.zero_grad()
        self.fake_H = self.netG(self.var_L)
        l_pix = self.l_pix_w * self.cri_pix(self.fake_H, self.real_H)
        with amp.scale_loss(l_pix, self.optimizer_G) as scale_loss:
            scale_loss.backward()
        self.optimizer_G.step()
        self.log_dict['l_pix'] = l_pix.item()

    def get_current_log(self):
        return self.log_dict

    def get_current_visuals(self, data, maskOn=True, need_HR=True):
        out_dict = OrderedDict()
        out_dict['LR'] = self.var_L.detach()[0, 0].float().cpu()
        out_dict['SR'] = self.fake_H.detach()[0, 0].float().cpu()
        if maskOn:
            # the way we contructed mask it is 1 x 1 x depth x width x height
            mask = data['mask'].detach().float().cpu()[0, 0, :]
            out_dict['SR'] *= mask
        if need_HR:
            out_dict['HR'] = self.real_H.detach().float().cpu()[0, 0, :]
            if maskOn:
                out_dict['HR'] *= mask
        return out_dict

    def print_network(self):
        s, n = self.get_network_description(self.netG)
        if isinstance(self.netG, nn.DataParallel):
            net_struc_str = '{} - {}'.format(self.netG.__class__.__name__,
                                             self.netG.module.__class__.__name__)
        else:
            net_struc_str = '{}'.format(self.netG.__class__.__name__)

        logger.info('Network G structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
        logger.info(s)

    def load(self):
        load_path_G = self.opt['path']['pretrain_model_G']
        if load_path_G is not None:
            logger.info('Loading pretrained model for G [{:s}] ...'.format(load_path_G))
            self.load_network(load_path_G, self.netG)

    def save(self, iter_step):
        self.save_network(self.netG, 'G', iter_step)



import dominate
from dominate.tags import *
import os


class HTML:
    def __init__(self, web_dir, title, image_subdir='', reflesh=0):
        self.title = title
        self.web_dir = web_dir
        # self.img_dir = os.path.join(self.web_dir, )
        self.img_subdir = image_subdir
        self.img_dir = os.path.join(self.web_dir, image_subdir)
        if not os.path.exists(self.web_dir):
            os.makedirs(self.web_dir)
        if not os.path.exists(self.img_dir):
            os.makedirs(self.img_dir)
        # print(self.img_dir)

        self.doc = dominate.document(title=title)
        if reflesh > 0:
            with self.doc.head:
                meta(http_equiv="reflesh", content=str(reflesh))

    def get_image_dir(self):
        return self.img_dir

    def add_header(self, str):
        with self.doc:
            h3(str)

    def add_table(self, border=1):
        self.t = table(border=border, style="table-layout: fixed;")
        self.doc.add(self.t)

    def add_images(self, ims, txts, links, width=400):
        self.add_table()
        with self.t:
            with tr():
                for im, txt, link in zip(ims, txts, links):
                    with td(style="word-wrap: break-word;", halign="center", valign="top"):
                        with p():
                            with a(href=os.path.join(link)):
                                img(style="width:%dpx" % width, src=os.path.join(im))
                            br()
                            p(txt)

    def save(self,file='index'):
        html_file = '%s/%s.html' % (self.web_dir,file)
        f = open(html_file, 'wt')
        f.write(self.doc.render())
        f.close()


if __name__ == '__main__':
    html = HTML('web/', 'test_html')
    html.add_header('hello world')

    ims = []
    txts = []
    links = []
    for n in range(4):
        ims.append('image_%d.png' % n)
        txts.append('text_%d' % n)
        links.append('image_%d.png' % n)
    html.add_images(ims, txts, links)
    html.save()



import torch

def train_collate(batch):
    batch_size = len(batch)
    inputs = torch.stack([batch[b][0] for b in range(batch_size)], 0)
    bboxes = [batch[b][1] for b in range(batch_size)]
    labels = [batch[b][2] for b in range(batch_size)]

    return [inputs, bboxes, labels]


def eval_collate(batch):
    batch_size = len(batch)
    inputs = torch.stack([batch[b][0]for b in range(batch_size)], 0)
    bboxes = [batch[b][1] for b in range(batch_size)]
    labels = [batch[b][2] for b in range(batch_size)]
    images = [batch[b][3] for b in range(batch_size)]

    return [inputs, bboxes, labels, images]


def test_collate(batch):
    batch_size = len(batch)
    for b in range(batch_size): 
        inputs = torch.stack([batch[b][0]for b in range(batch_size)], 0)
        images = [batch[b][1] for b in range(batch_size)]

    return [inputs, images]



import nrrd
import bm3d
import numpy as np
from skimage.restoration import estimate_sigma
import os
import h5py
import json
from tqdm import tqdm
import torch 


def rescale_intensity(img):
    img = (img - img.min()) / (img.max() - img.min())
    return img

def rescale_img(img, clip=None, max_val=255.0):
    if clip:
        img = np.clip(img, clip[0], clip[1])
    img = rescale_intensity(img)
    img *= max_val
    img = img.astype(np.float32)
    # img = img.type(torch.float32)
    return img

def read_config(config_path):
    config_path = os.path.join(config_path) 
    with open(config_path, 'r') as f:
        config = json.load(f)
    meta_data = {}
    meta_data['Spacing'] = [float(i) for i in config['Spacing'].split()]
    meta_data['Orientation'] = [float(i) for i in config['Orientation'].split()] 
    meta_data['Origin'] = [float(i) for i in config['Origin'].split()]
    header = {'units': ['mm', 'mm', 'mm'], 'spacings': meta_data['Spacing']} 
    return header


def calculate_bm3d(lr_scan_path, UID, save_path, split_uid_for_ref=False):
    with open(UID, 'r') as f:
        lines = f.readlines()
        uids = [l.rstrip() for l in lines]
    if not os.path.exists(save_path):
        os.makedirs(save_path)
    
    for uid in uids:
        print('Reading UID:', uid)
        LR_path = os.path.join(lr_scan_path, '{}.h5'.format(uid))
        with h5py.File(LR_path, 'r') as file:
            lr_scan = file['data'][()]

        lr_bm3d_norm = np.empty(lr_scan.shape)
        header = os.path.join(lr_scan_path, '{}.json'.format(uid))
        volume_save_path = os.path.join(save_path, '{}.nrrd'.format(uid))
        spacing_info = read_config(header)

        for index, slice_ in enumerate(tqdm(lr_scan)):
            # +1000 to input images 
            # 
            img_slice = rescale_img(slice_, clip=None, max_val=255.0)
            sigma_est = np.mean(estimate_sigma(img_slice, multichannel=False))
            BM3D_denoised_image = bm3d.bm3d(img_slice, sigma_psd=sigma_est, stage_arg=bm3d.BM3DStages.ALL_STAGES) # bm3d.BM3DStages.HARD_THRESHOLDING
            bm3d_clipped = rescale_img(BM3D_denoised_image, clip=(0.0, 255.0), max_val=slice_.max())
            lr_bm3d_norm[index] = bm3d_clipped
        
        lr_bm3d_norm = lr_bm3d_norm.astype(np.int16)
        lr_bm3d_norm_clipped = np.clip(lr_bm3d_norm, 0, 1500)
        lr_bm3d_norm_clipped = lr_bm3d_norm_clipped.round()
        lr_bm3d_norm_clipped = lr_bm3d_norm_clipped.astype(np.int16)
        nrrd.write(volume_save_path, lr_bm3d_norm_clipped, spacing_info, index_order='C')

"""
Apply bm3d denoising to data['LR'] of shape 3D(D,H,W) and value [0,1]
data: dictionary from dataloader 
"""
def calculate_scan_bm3d(data):
    lr_scan = data['LR'][0,0].numpy()
    lr_bm3d_norm = np.empty(lr_scan.shape)
    for index, slice_ in enumerate(lr_scan):
        img_slice = rescale_img(slice_, clip=None, max_val=255.0)
        sigma_est = np.mean(estimate_sigma(img_slice, multichannel=False))
        BM3D_denoised_image = bm3d.bm3d(img_slice, sigma_psd=sigma_est, stage_arg=bm3d.BM3DStages.ALL_STAGES) # bm3d.BM3DStages.HARD_THRESHOLDING
        # bm3d_clipped = rescale_img(BM3D_denoised_image, clip=(0.0, 255.0), max_val=slice_.max())
        bm3d_clipped = rescale_img(BM3D_denoised_image, clip=(0.0, 255.0), max_val=1500)
        lr_bm3d_norm[index] = bm3d_clipped
    
    # lr_bm3d_norm = lr_bm3d_norm.astype(np.int16)
    lr_bm3d_norm_clipped = np.clip(lr_bm3d_norm, 0, 1500)
    lr_bm3d_norm_clipped = lr_bm3d_norm_clipped.round()
    # lr_bm3d_norm_clipped = lr_bm3d_norm_clipped.astype(np.int16)
    data['volume'] = torch.from_numpy(lr_bm3d_norm_clipped)
    print(lr_bm3d_norm[10])
    return data

if __name__ == '__main__':
    dataroot_LR = '/datasets/data_st1.0_merged'
    UID = '/datasets/uids/70-30_split/ucla_testUID_70_30_split.txt'
    save_path = '../../results/BM3D-70-30-Split_tileStitch-32x64x64_Ref-k2d100/merged_cases/test_st1.0_norm'
    calculate_bm3d(dataroot_LR, UID, save_path)


from common import *
from net.lib.box.overlap.cython_box_overlap import cython_box_overlap
from net.lib.box.nms.torch_nms import torch_nms
from net.lib.box.nms.gpu_nms import gpu_nms
from net.lib.box.nms.cython_nms import cython_nms

#from model.lib.box.nms.py_nms import py_nms
#


## torch #####################################################################

# Clip process to image boundaries.
def torch_clip_boxes(boxes, width, height):
    boxes = torch.stack(
        (boxes[:,0].clamp(0, width  - 1),
         boxes[:,1].clamp(0, height - 1),
         boxes[:,2].clamp(0, width  - 1),
         boxes[:,3].clamp(0, height - 1)), 1)

    return boxes


def torch_box_transform(boxes, targets):
  bw = boxes[:, 2] - boxes[:, 0] + 1.0
  bh = boxes[:, 3] - boxes[:, 1] + 1.0
  bx = boxes[:, 0] + 0.5 * bw
  by = boxes[:, 1] + 0.5 * bh

  tw = targets[:, 2] - targets[:, 0] + 1.0
  th = targets[:, 3] - targets[:, 1] + 1.0
  tx = targets[:, 0] + 0.5 * tw
  ty = targets[:, 1] + 0.5 * th

  dx = (tx - bx) / bw
  dy = (ty - by) / bh
  dw = torch.log(tw / bw)
  dh = torch.log(th / bh)

  deltas = torch.stack((dx, dy, dw, dh), 1)
  return deltas


def torch_box_transform_inv(boxes, deltas):

    bw = boxes[:, 2] - boxes[:, 0] + 1.0
    bh = boxes[:, 3] - boxes[:, 1] + 1.0
    bx = boxes[:, 0] + 0.5 * bw
    by = boxes[:, 1] + 0.5 * bh
    bw = bw.unsqueeze(1)
    bh = bh.unsqueeze(1)
    bx = bx.unsqueeze(1)
    by = by.unsqueeze(1)

    dx = deltas[:, 0::4]
    dy = deltas[:, 1::4]
    dw = deltas[:, 2::4]
    dh = deltas[:, 3::4]

    x = bx + dx * bw
    y = by + dy * bh
    w = torch.exp(dw) * bw
    h = torch.exp(dh) * bh

    predictions = torch.cat( (
        x - 0.5 * w,
        y - 0.5 * h,
        x + 0.5 * w,
        y + 0.5 * h), 1)

    return predictions


def torch_filter_boxes(boxes, min_size):
    ws = boxes[:, 2] - boxes[:, 0] + 1
    hs = boxes[:, 3] - boxes[:, 1] + 1
    keep = (((ws >= min_size) + (hs >= min_size))==2).nonzero().view(-1)
    return keep


def torch_box_overlap(boxes, gt_boxes):

    '''
    Parameters
    ----------
    boxes: (N, 4) ndarray or tensor or variable
    query_boxes: (K, 4) ndarray or tensor or variable
    Returns
    -------
    overlaps: (N, K) overlap between boxes and gt_boxes
    '''

    box_areas = (boxes   [:, 2] - boxes   [:, 0] + 1) *  (boxes   [:, 3] - boxes   [:, 1] + 1)
    gt_areas  = (gt_boxes[:, 2] - gt_boxes[:, 0] + 1) *  (gt_boxes[:, 3] - gt_boxes[:, 1] + 1)

    intersect_ws = (torch.min(boxes[:, 2:3], gt_boxes[:, 2:3].t()) - torch.max(boxes[:, 0:1], gt_boxes[:, 0:1].t()) + 1).clamp(min=0)
    intersect_hs = (torch.min(boxes[:, 3:4], gt_boxes[:, 3:4].t()) - torch.max(boxes[:, 1:2], gt_boxes[:, 1:2].t()) + 1).clamp(min=0)
    intersect_areas = intersect_ws * intersect_hs
    union_areas     = box_areas.view(-1, 1) + gt_areas.view(1, -1) - intersect_areas
    overlaps        = intersect_areas / union_areas

    return overlaps


# def torch_clip_boxes(boxes, width, height):
#     ''' Clip process to image boundaries. '''
#
#     boxes = torch.stack(
#         (boxes[:,0].clamp(0, width  - 1),
#          boxes[:,1].clamp(0, height - 1),
#          boxes[:,2].clamp(0, width  - 1),
#          boxes[:,3].clamp(0, height - 1)), 1)
#
#     return boxes





## python  #####################################################################

# # overlaps
# #https://stackoverflow.com/questions/25349178/calculating-percentage-of-bounding-box-overlap-for-image-detector-evaluation
# #http://www.pyimagesearch.com/2016/11/07/intersection-over-union-iou-for-object-detection/
#
# def one_box_overlap(box1,box2):
#
#     ''' Calculate the Intersection over Union (IoU) of two bounding boxes.
#
#     '''
#     #box=x0,y0,x1,y1
#     assert box1[0] < box1[2]
#     assert box1[1] < box1[3]
#     assert box2[0] < box2[2]
#     assert box2[1] < box2[3]
#
#     # determine the coordinates of the intersection rectangle
#     x_left   = max(box1[0], box2[0])
#     y_top    = max(box1[1], box2[1])
#     x_right  = min(box1[2], box2[2])
#     y_bottom = min(box1[3], box2[3])
#
#     if x_right < x_left or y_bottom < y_top:
#         return 0.0
#
#     # The intersection of two axis-aligned bounding boxes is always an
#     # axis-aligned bounding box
#     intersection_area = (x_right - x_left+1) * (y_bottom - y_top+1)
#
#     # compute the area of both AABBs
#     area1 = (box1[2] - box1[0]+1) * (box1[3] - box1[1]+1)
#     area2 = (box2[2] - box2[0]+1) * (box2[3] - box2[1]+1)
#
#     # compute the intersection over union by taking the intersection
#     # area and dividing it by the sum of prediction + ground-truth
#     # areas - the interesection area
#     iou = intersection_area / float(area1 + area2 - intersection_area)
#     assert iou >= 0.0
#     assert iou <= 1.0
#     return iou
#
#


def box_overlap(boxes, gt_boxes):

    box_areas = (boxes   [:, 2] - boxes   [:, 0] + 1) *  (boxes   [:, 3] - boxes   [:, 1] + 1)
    gt_areas  = (gt_boxes[:, 2] - gt_boxes[:, 0] + 1) *  (gt_boxes[:, 3] - gt_boxes[:, 1] + 1)

    intersect_ws = np.clip((np.min(boxes[:, 2:3], gt_boxes[:, 2:3].t()) - np.max(boxes[:, 0:1], gt_boxes[:, 0:1].t()) + 1),0,1e8)
    intersect_hs = np.clip((np.min(boxes[:, 3:4], gt_boxes[:, 3:4].t()) - np.max(boxes[:, 1:2], gt_boxes[:, 1:2].t()) + 1),0,1e8)
    intersect_areas = intersect_ws * intersect_hs
    union_areas     = box_areas.view(-1, 1) + gt_areas.view(1, -1) - intersect_areas
    overlaps        = intersect_areas / union_areas

    return overlaps





## python  ##############################################################################
def clip_boxes(boxes, width, height):
    boxes[:, 0] = np.clip(boxes[:, 0], 0, width  - 1)
    boxes[:, 1] = np.clip(boxes[:, 1], 0, height - 1)
    boxes[:, 2] = np.clip(boxes[:, 2], 0, width  - 1)
    boxes[:, 3] = np.clip(boxes[:, 3], 0, height - 1)
    return boxes


def box_transform(windows, targets):
    # targets : ground truth

    bw = windows[:, 2] - windows[:, 0] + 1.0
    bh = windows[:, 3] - windows[:, 1] + 1.0
    bx = windows[:, 0] + 0.5 *bw
    by = windows[:, 1] + 0.5 *bh

    tw = targets[:, 2] - targets[:, 0] + 1.0
    th = targets[:, 3] - targets[:, 1] + 1.0
    tx = targets[:, 0] + 0.5 * tw
    ty = targets[:, 1] + 0.5 * th

    dx = (tx - bx) / bw
    dy = (ty - by) / bh
    dw = np.log(tw / bw)
    dh = np.log(th / bh)

    deltas = np.vstack((dx, dy, dw, dh)).transpose()
    return deltas


def box_transform_inv(windows, deltas):

    num   = len(windows)
    predictions = np.zeros((num,4), dtype=np.float32)
    # if num == 0: return predictions  #not possible?

    bw = windows[:, 2] - windows[:, 0] + 1.0
    bh = windows[:, 3] - windows[:, 1] + 1.0
    bx = windows[:, 0] + 0.5 * bw
    by = windows[:, 1] + 0.5 * bh
    bw = bw[:, np.newaxis]
    bh = bh[:, np.newaxis]
    bx = bx[:, np.newaxis]
    by = by[:, np.newaxis]

    dx = deltas[:, 0::4]
    dy = deltas[:, 1::4]
    dw = deltas[:, 2::4]
    dh = deltas[:, 3::4]

    x = dx * bw + bx
    y = dy * bh + by
    w = np.exp(dw) * bw
    h = np.exp(dh) * bh

    predictions[:, 0::4] = x - 0.5 * w  # x0,y0,x1,y1
    predictions[:, 1::4] = y - 0.5 * h
    predictions[:, 2::4] = x + 0.5 * w
    predictions[:, 3::4] = y + 0.5 * h

    return predictions


def filter_boxes(boxes, min_size):
    ws = boxes[:, 2] - boxes[:, 0] + 1
    hs = boxes[:, 3] - boxes[:, 1] + 1
    keep = np.where((ws >= min_size) & (hs >= min_size))[0]
    return keep

def is_small_box_at_boundary(box,W,H, min_size):
    x0,y0,x1,y1 = box
    w = (x1-x0)+1
    h = (y1-y0)+1
    aspect = max(w,h)/min(w,h)
    area = w*h
    return ((x0==0 or x1==W-1) and (w<min_size)) or \
           ((y0==0 or y1==H-1) and (h<min_size))



def is_small_box(box, min_size):
    x0,y0,x1,y1 = box
    w = (x1-x0)+1
    h = (y1-y0)+1
    aspect = max(w,h)/min(w,h)
    area = w*h
    return (w <min_size or h<min_size)

def is_big_box(box, max_size):
    x0,y0,x1,y1 = box
    w = (x1-x0)+1
    h = (y1-y0)+1
    aspect = max(w,h)/min(w,h)
    area = w*h
    return (w>max_size or h>max_size)





## check  ##############################################################################

def run_check_nms():

    #test nms:
    H,W = 480,640
    num_objects = 4
    rois = []
    for n in range(num_objects):
        w = np.random.randint(64,256)
        h = np.random.randint(64,256)
        x0 = np.random.randint(0,W-w)
        y0 = np.random.randint(0,H-h)
        x1 = x0 + w
        y1 = y0 + h
        gt = [x0,y0,x1,y1]

        M = np.random.randint(10,20)
        for m in range(M):
            dw = int(np.random.uniform(0.5,2)*w)
            dh = int(np.random.uniform(0.5,2)*h)
            dx = int(np.random.uniform(-1,1)*w*0.5)
            dy = int(np.random.uniform(-1,1)*h*0.5)
            xx0 = x0 - dw//2 + dx
            yy0 = y0 - dh//2 + dy
            xx1 = xx0 + w+dw
            yy1 = yy0 + h+dh
            score = np.random.uniform(0.5,2)

            rois.append([xx0,yy0,xx1,yy1, score])
            pass

    rois = np.array(rois).astype(np.float32)


    if 1:
        keep = gpu_nms(rois, 0.5)
        print('gpu_nms    :', keep)
        keep = cython_nms(rois, 0.5)
        print('cython_nms :', keep)

        #gpu     [52, 39, 48, 43, 47, 21, 9, 6, 32, 8, 16, 36, 28, 29, 53, 41]
        #py      [52, 39, 48, 43, 47, 21, 9, 6, 32, 8, 16, 36, 28, 29, 53, 41]
        #cython  [52, 39, 48, 43, 47, 21, 9, 6, 32, 8, 16, 36, 28, 29, 53, 41]

    if 1:
        rois = torch.from_numpy(rois).cuda()
        keep = torch_nms(rois, 0.5)

        keep = keep.cpu().numpy()
        print('torch_nms  :', keep)
        #torch [52 39 48 43 47 21  9  6 32  8 16 36 28 29 53 41]


# # main #################################################################
if __name__ == '__main__':
    print( '%s: calling main function ... ' % os.path.basename(__file__))

    run_check_nms()

    print('succcess!')



from setting import parse_opts 
from model import generate_model
import torch
from torch import nn
import numpy as np
from torch.utils.data import DataLoader
import torch.nn.functional as F
import sys
import os
import numpy as np
from tqdm import tqdm
from utils.logger import log
from build_unet_model import UNet
import nrrd


def getIoU(y_true, y_pred):
    y_true_f = y_true.flatten()
    y_pred_f = y_pred.flatten()
    intersection = np.logical_and(y_true_f, y_pred_f).sum()
    union = np.logical_or(y_true_f, y_pred_f).sum()
    return (intersection + 1) * 1. / (union + 1)

def rescale_img_to_rgb(img):
    img = (img - img.min()) / (img.max() - img.min())
    img *= 255.0
    img = img.astype(np.uint8)
    return img


if __name__ == '__main__':
    # settting
    sets = parse_opts()
    path_to_data = '/datasets'
    save_output = '/datasets/Unet_Predictions'

    os.environ["CUDA_VISIBLE_DEVICES"]=str(sets.gpu_id[0])
    # ---------------------------------------------------   
    model = UNet(in_nc=1, out_nc=1, nf=64).cuda()
    model = nn.DataParallel(model)
    path_to_weight = './trails_UNET/models/UNet_epoch_106_best_weight_0.646983.pth.tar'
    checkpoint = torch.load(path_to_weight)
    model.load_state_dict(checkpoint['state_dict'])
    print('UNet weights loaded!')

    # run inference
    with torch.no_grad():
        model.eval()
        norm_models = [norm for norm in os.listdir(path_to_data) if norm != 'Unet_Predictions']
        print('norm models:', norm_models)

        for norm_model in norm_models:
            print('Loading Normalization:', norm_model)
            save_model_predictions = os.path.join(save_output, norm_model)
            print(save_model_predictions)
            if not os.path.exists(save_model_predictions):
                os.makedirs(save_model_predictions)

            cur_path = os.path.join(path_to_data, norm_model)
            cnds =  os.listdir(cur_path)
            for each_cnd in cnds:
                save_model_cnd = os.path.join(save_model_predictions, each_cnd)
                if not os.path.exists(save_model_cnd):
                    os.makedirs(save_model_cnd)
                volumes = os.path.join(cur_path, each_cnd)

                for vol in os.listdir(volumes):
                    path_to_vol = os.path.join(volumes, vol)
                    data, header = nrrd.read(path_to_vol)
                    mask_header = {'units': ['mm', 'mm', 'mm'], 'spacings': header['spacings']}
                    rescaled_vol = rescale_img_to_rgb(data)
                    rescaled_vol = (rescaled_vol - rescaled_vol.mean()) / rescaled_vol.std()
                    rescaled_vol = torch.from_numpy(np.expand_dims(np.expand_dims(rescaled_vol.transpose(2,0,1), axis=0), axis=0)).type(torch.FloatTensor)

                    # get prediction
                    rescaled_vol = rescaled_vol.cuda() 
                    out_masks = model(rescaled_vol)
                    output_prob = torch.sigmoid(out_masks).detach().cpu().numpy()
                    output_prob_thresh = (output_prob > 0.5) * 1
                    vol_to_save = output_prob_thresh[0,0,:,:,:]
                    vol_to_save = vol_to_save*255.0
                    vol_to_save = vol_to_save.astype(np.uint8)
                    vol_save_path = os.path.join(save_model_cnd, vol)
                    nrrd.write(vol_save_path, vol_to_save, mask_header, index_order='C')
            
    print('All Normalization Prediction Complete!')









import numbers
import numpy as np
from sklearn.utils import check_array, check_random_state
from numpy.lib.stride_tricks import as_strided
from itertools import product

def generate_test_locations(patch_size, stride, im_size):
    stride_size_x = patch_size[0] / stride[0]
    stride_size_y = patch_size[1] / stride[1]
    stride_size_z = patch_size[2] / stride[2]
    pad_x = (int(patch_size[0] / 2), int(np.ceil(im_size[0] / stride_size_x) * stride_size_x - im_size[0] + patch_size[0] / 2))
    pad_y = (int(patch_size[1] / 2), int(np.ceil(im_size[1] / stride_size_y) * stride_size_y - im_size[1] + patch_size[1] / 2))
    pad_z = (int(patch_size[2] / 2), int(np.ceil(im_size[2] / stride_size_z) * stride_size_z - im_size[2] + patch_size[2] / 2))
    x = np.arange(patch_size[0] / 2, im_size[0] + pad_x[0] + pad_x[1] - patch_size[0] / 2 + 1, stride_size_x)
    y = np.arange(patch_size[1] / 2, im_size[1] + pad_y[0] + pad_y[1] - patch_size[1] / 2 + 1, stride_size_y)
    z = np.arange(patch_size[2] / 2, im_size[2] + pad_z[0] + pad_z[1] - patch_size[2] / 2 + 1, stride_size_z)
    return (x, y, z), (pad_x, pad_y, pad_z)


def zero_padding(vol, window_size, overlap=0.5):
    """
    vol and window size: DxWxH 
    """
    for i in range(3): 
        if not (window_size[i]*overlap).is_integer(): 
            raise ValueError('Window overlap size not supported')
    overlap = np.asarray(window_size) * overlap #[16x32x32]
    pad_z = overlap[0] - (vol.shape[0] % overlap[0])
    # Assume vol is a 3D numpy array with shape (n, 512, 512)
    pad_width = ((0, int(pad_z)), (0, 0), (0, 0))
    vol_padded = np.pad(vol, pad_width=pad_width, mode='constant', constant_values=0)
    return vol_padded 

def remove_padding(pad_result, image):
    z, y, x = image.shape
    return pad_result[:z, :y, :x]

"""
@author: konop
Support for extracting 3d patches similar to extract patches 2d from scikit-learn
source: https://github.com/konopczynski/Vessel3DDL
"""

"""
Arrays used for testing

a = np.array([[[ 1, 2, 3, 4],[ 5, 6, 7, 8],[ 9,10,11,12],[13,14,15,16]],
              [[17,18,19,20],[21,22,23,24],[25,26,27,28],[29,30,31,32]],
              [[33,34,35,36],[37,38,39,40],[41,42,43,44],[45,46,47,48]],
              [[49,50,51,52],[53,54,55,56],[57,58,59,60],[61,62,63,64]]])

i = np.array([[ 1, 2, 3, 4],
              [ 5, 6, 7, 8],
              [ 9,10,11,12],
              [13,14,15,16]])

i_modal = np.array([[[ 1, 2, 3],[ 4, 5, 6],[ 7, 8, 9],[10,11,12]],
                    [[13,14,15],[16,17,18],[19,20,21],[22,23,24]],
                    [[25,26,27],[28,29,30],[31,32,33],[34,35,36]],
                    [[37,38,39],[40,41,42],[43,44,45],[46,46,48]]])

a_modal = np.array([[[[  1,  2,  3],[  4,  5,  6],[  7,  8,  9],[ 10, 11, 12]],
                     [[ 13, 14, 15],[ 16, 17, 18],[ 19, 20, 21],[ 22, 23, 24]],
                     [[ 25, 26, 27],[ 28, 29, 30],[ 31, 32, 33],[ 34, 35, 36]],
                     [[ 37, 38, 39],[ 40, 41, 42],[ 43, 44, 45],[ 46, 47, 48]]],
                    [[[ 49, 50, 51],[ 52, 53, 54],[ 55, 56, 57],[ 58, 59, 60]],
                     [[ 61, 62, 63],[ 64, 65, 66],[ 67, 68, 69],[ 70, 71, 72]],
                     [[ 73, 74, 75],[ 76, 77, 78],[ 79, 80, 81],[ 82, 83, 84]],
                     [[ 85, 86, 87],[ 88, 89, 90],[ 91, 92, 93],[ 94, 95, 96]]],
                    [[[ 97, 98, 99],[100,101,102],[103,104,105],[106,107,108]],
                     [[109,110,111],[112,113,114],[115,116,117],[118,119,120]],
                     [[121,122,123],[124,125,126],[127,128,129],[130,131,132]],
                     [[133,134,135],[136,137,138],[139,140,141],[142,143,144]]],
                    [[[145,146,147],[148,149,150],[151,152,153],[154,155,156]],
                     [[157,158,159],[160,161,162],[163,164,165],[166,167,168]],
                     [[169,170,171],[172,173,174],[175,176,177],[178,179,180]],
                     [[181,182,183],[184,185,186],[187,188,189],[190,191,192]]]])

i_modal = np.array([[[ 1, 2],[ 4, 5],[ 7, 8],[10,11]],
                    [[13,14],[16,17],[19,20],[22,23]],
                    [[25,26],[28,29],[31,32],[34,35]],
                    [[37,38],[40,41],[43,44],[46,46]]])

a_modal = np.array([[[[  1,  2],[  4,  5],[  7,  8],[ 10, 11]],
                     [[ 13, 14],[ 16, 17],[ 19, 20],[ 22, 23]],
                     [[ 25, 26],[ 28, 29],[ 31, 32],[ 34, 35]],
                     [[ 37, 38],[ 40, 41],[ 43, 44],[ 46, 47]]],
                    [[[ 49, 50],[ 52, 53],[ 55, 56],[ 58, 59]],
                     [[ 61, 62],[ 64, 65],[ 67, 68],[ 70, 71]],
                     [[ 73, 74],[ 76, 77],[ 79, 80],[ 82, 83]],
                     [[ 85, 86],[ 88, 89],[ 91, 92],[ 94, 95]]],
                    [[[ 97, 98],[100,101],[103,104],[106,107]],
                     [[109,110],[112,113],[115,116],[118,119]],
                     [[121,122],[124,125],[127,128],[130,131]],
                     [[133,134],[136,137],[139,140],[142,143]]],
                    [[[145,146],[148,149],[151,152],[154,155]],
                     [[157,158],[160,161],[163,164],[166,167]],
                     [[169,170],[172,173],[175,176],[178,179]],
                     [[181,182],[184,185],[187,188],[190,191]]]])
"""


def _compute_n_patches(i_h, i_w, p_h, p_w, max_patches=None):
    """Compute the number of patches that will be extracted in an image.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    i_h : int
        The image height
    i_w : int
        The image with
    p_h : int
        The height of a patch
    p_w : int
        The width of a patch
    max_patches : integer or float, optional default is None
        The maximum number of patches to extract. If max_patches is a float
        between 0 and 1, it is taken to be a proportion of the total number
        of patches.
    """
    n_h = i_h - p_h + 1
    n_w = i_w - p_w + 1
    all_patches = n_h * n_w

    if max_patches:
        if (isinstance(max_patches, (numbers.Integral))
            and max_patches < all_patches):
            return max_patches
        elif (isinstance(max_patches, (numbers.Real))
              and 0 < max_patches < 1):
            return int(max_patches * all_patches)
        else:
            raise ValueError("Invalid value for max_patches: %r" % max_patches)
    else:
        return all_patches


def _compute_n_patches_3d(i_x, i_y, i_z, p_x, p_y, p_z, max_patches=None):
    """Compute the number of patches that will be extracted in a volume.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    i_x : int
        The number of voxels in x dimension
    i_y : int
        The number of voxels in y dimension
    i_z : int
        The number of voxels in z dimension
    p_x : int
        The number of voxels in x dimension of a patch
    p_y : int
        The number of voxels in y dimension of a patch
    p_z : int
        The number of voxels in z dimension of a patch
    max_patches : integer or float, optional default is None
        The maximum number of patches to extract. If max_patches is a float
        between 0 and 1, it is taken to be a proportion of the total number
        of patches.
    """
    n_x = i_x - p_x + 1
    n_y = i_y - p_y + 1
    n_z = i_z - p_z + 1
    all_patches = n_x * n_y * n_z

    if max_patches:
        if (isinstance(max_patches, (numbers.Integral))
            and max_patches < all_patches):
            return max_patches
        elif (isinstance(max_patches, (numbers.Real))
              and 0 < max_patches < 1):
            return int(max_patches * all_patches)
        else:
            raise ValueError("Invalid value for max_patches: %r" % max_patches)
    else:
        return all_patches


def extract_patches(arr, patch_shape=8, extraction_step=1):
    """Extracts patches of any n-dimensional array in place using strides.
    Given an n-dimensional array it will return a 2n-dimensional array with
    the first n dimensions indexing patch position and the last n indexing
    the patch content. This operation is immediate (O(1)). A reshape
    performed on the first n dimensions will cause numpy to copy data, leading
    to a list of extracted patches.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    arr : ndarray
        n-dimensional array of which patches are to be extracted
    patch_shape : integer or tuple of length arr.ndim
        Indicates the shape of the patches to be extracted. If an
        integer is given, the shape will be a hypercube of
        sidelength given by its value.
    extraction_step : integer or tuple of length arr.ndim
        Indicates step size at which extraction shall be performed.
        If integer is given, then the step is uniform in all dimensions.
    Returns
    -------
    patches : strided ndarray
        2n-dimensional array indexing patches on first n dimensions and
        containing patches on the last n dimensions. These dimensions
        are fake, but this way no data is copied. A simple reshape invokes
        a copying operation to obtain a list of patches:
        result.reshape([-1] + list(patch_shape))
    """

    arr_ndim = arr.ndim

    if isinstance(patch_shape, numbers.Number):
        patch_shape = tuple([patch_shape] * arr_ndim)
    if isinstance(extraction_step, numbers.Number):
        extraction_step = tuple([extraction_step] * arr_ndim)

    patch_strides = arr.strides

    slices = [slice(None, None, st) for st in extraction_step]
    indexing_strides = arr[slices].strides

    patch_indices_shape = ((np.array(arr.shape) - np.array(patch_shape)) //
                           np.array(extraction_step)) + 1

    shape = tuple(list(patch_indices_shape) + list(patch_shape))
    strides = tuple(list(indexing_strides) + list(patch_strides))

    patches = as_strided(arr, shape=shape, strides=strides)
    return patches


def extract_patches_2d(image, patch_size, max_patches=None, random_state=None):
    """Reshape a 2D image into a collection of patches
    The resulting patches are allocated in a dedicated array.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    image : array, shape = (image_height, image_width) or
        (image_height, image_width, n_channels)
        The original image data. For color images, the last dimension specifies
        the channel: a RGB image would have `n_channels=3`.
    patch_size : tuple of ints (patch_height, patch_width)
        the dimensions of one patch
    max_patches : integer or float, optional default is None
        The maximum number of patches to extract. If max_patches is a float
        between 0 and 1, it is taken to be a proportion of the total number
        of patches.
    random_state : int or RandomState
        Pseudo number generator state used for random sampling to use if
        `max_patches` is not None.
    Returns
    -------
    patches : array, shape = (n_patches, patch_height, patch_width) or
         (n_patches, patch_height, patch_width, n_channels)
         The collection of patches extracted from the image, where `n_patches`
         is either `max_patches` or the total number of patches that can be
         extracted.
    Examples
    --------
    # >>> from sklearn.feature_extraction import image
    # >>> one_image = np.arange(16).reshape((4, 4))
    # >>> one_image
    array([[ 0,  1,  2,  3],
           [ 4,  5,  6,  7],
           [ 8,  9, 10, 11],
           [12, 13, 14, 15]])
    # >>> patches = image.extract_patches_2d(one_image, (2, 2))
    # >>> print(patches.shape)
    (9, 2, 2)
    # >>> patches[0]
    array([[0, 1],
           [4, 5]])
    # >>> patches[1]
    array([[1, 2],
           [5, 6]])
    # >>> patches[8]
    array([[10, 11],
           [14, 15]])
    """
    i_h, i_w = image.shape[:2]
    p_h, p_w = patch_size

    if p_h > i_h:
        raise ValueError("Height of the patch should be less than the height"
                         " of the image.")

    if p_w > i_w:
        raise ValueError("Width of the patch should be less than the width"
                         " of the image.")

    image = check_array(image, allow_nd=True)
    image = image.reshape((i_h, i_w, -1))
    n_colors = image.shape[-1]

    extracted_patches = extract_patches(image, patch_shape=(p_h, p_w, n_colors), extraction_step=1)

    n_patches = _compute_n_patches(i_h, i_w, p_h, p_w, max_patches)
    if max_patches:
        rng = check_random_state(random_state)
        i_s = rng.randint(i_h - p_h + 1, size=n_patches)
        j_s = rng.randint(i_w - p_w + 1, size=n_patches)
        patches = extracted_patches[i_s, j_s, 0]
    else:
        patches = extracted_patches

    patches = patches.reshape(-1, p_h, p_w, n_colors)
    # remove the color dimension if useless
    if patches.shape[-1] == 1:
        return patches.reshape((n_patches, p_h, p_w))
    else:
        return patches


def extract_patches_3d(volume, patch_size, max_patches=None, random_state=None):
    """Reshape a 3D volume into a collection of patches
    The resulting patches are allocated in a dedicated array.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    volume : array, shape = (volume_x, volume_y, volume_z)
        No channels are allowed
    patch_size : tuple of ints (patch_x, patch_y, patch_z)
        the dimensions of one patch
    max_patches : integer or float, optional default is None
        The maximum number of patches to extract. If max_patches is a float
        between 0 and 1, it is taken to be a proportion of the total number
        of patches.
    random_state : int or RandomState
        Pseudo number generator state used for random sampling to use if
        `max_patches` is not None.
    Returns
    -------
    patches : array, shape = (n_patches, patch_x, patch_y, patch_z)
         The collection of patches extracted from the volume, where `n_patches`
         is either `max_patches` or the total number of patches that can be
         extracted.
    Examples
    --------
    TBD
    """
    v_x, v_y, v_z = volume.shape[:3]
    p_x, p_y, p_z = patch_size

    if p_x > v_x:
        raise ValueError("Height of the patch should be less than the height"
                         " of the volume.")

    if p_y > v_y:
        raise ValueError("Width of the patch should be less than the width"
                         " of the volume.")

    if p_z > v_z:
        raise ValueError("z of the patch should be less than the z"
                         " of the volume.")

    volume = check_array(volume, allow_nd=True)
    volume = volume.reshape((v_x, v_y, v_z, -1))
    n_colors = volume.shape[-1]

    extracted_patches = extract_patches(volume, patch_shape=(p_x, p_y, p_z, n_colors), extraction_step=1)

    n_patches = _compute_n_patches_3d(v_x, v_y, v_z, p_x, p_y, p_z, max_patches)
    if max_patches:
        rng = check_random_state(random_state)
        i_s = rng.randint(v_x - p_x + 1, size=n_patches)
        j_s = rng.randint(v_y - p_y + 1, size=n_patches)
        k_s = rng.randint(v_z - p_z + 1, size=n_patches)

        patches = extracted_patches[i_s, j_s, k_s, 0]
    else:
        patches = extracted_patches

    patches = patches.reshape(-1, p_x, p_y, p_z, n_colors)
    # remove the color dimension if useless
    if patches.shape[-1] == 1:
        return patches.reshape((n_patches, p_x, p_y, p_z))
    else:
        return patches


def extract_patches_3d_fromMask(volume, mask, patch_size, max_patches=None, random_state=None):
    """Reshape a 3D volume into a collection of patches
    The resulting patches are allocated in a dedicated array.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    volume : array, shape = (volume_x, volume_y, volume_z)
        No channels are allowed
    patch_size : tuple of ints (patch_x, patch_y, patch_z)
        the dimensions of one patch
    max_patches : integer or float, optional default is None
        The maximum number of patches to extract. If max_patches is a float
        between 0 and 1, it is taken to be a proportion of the total number
        of patches.
    random_state : int or RandomState
        Pseudo number generator state used for random sampling to use if
        `max_patches` is not None.
    Returns
    -------
    patches : array, shape = (n_patches, patch_x, patch_y, patch_z)
         The collection of patches extracted from the volume, where `n_patches`
         is either `max_patches` or the total number of patches that can be
         extracted.
    Examples
    --------
    TBD
    """
    v_x, v_y, v_z = volume.shape[:3]
    p_x, p_y, p_z = patch_size

    if p_x > v_x:
        raise ValueError("Height of the patch should be less than the height"
                         " of the volume.")

    if p_y > v_y:
        raise ValueError("Width of the patch should be less than the width"
                         " of the volume.")

    if p_z > v_z:
        raise ValueError("z of the patch should be less than the z"
                         " of the volume.")

    volume = check_array(volume, allow_nd=True)
    volume = volume.reshape((v_x, v_y, v_z, -1))
    n_colors = volume.shape[-1]

    extracted_patches = extract_patches(volume, patch_shape=(p_x, p_y, p_z, n_colors), extraction_step=1)

    n_patches = _compute_n_patches_3d(v_x, v_y, v_z, p_x, p_y, p_z, max_patches)
    # check the indexes where mask is True
    M = np.array(np.where(mask[p_x / 2:v_x - p_x / 2,
                          p_y / 2:v_y - p_y / 2,
                          p_z / 2:v_z - p_z / 2] == True)).T
    if max_patches:
        rng = check_random_state(random_state)
        indx = rng.randint(len(M), size=n_patches)
        i_s = M[indx][:, 0]
        j_s = M[indx][:, 1]
        k_s = M[indx][:, 2]
        patches = extracted_patches[i_s, j_s, k_s, 0]
    else:
        patches = extracted_patches

    patches = patches.reshape(-1, p_x, p_y, p_z, n_colors)
    # remove the color dimension if useless
    if patches.shape[-1] == 1:
        return patches.reshape((n_patches, p_x, p_y, p_z))
    else:
        return patches


def reconstruct_from_patches_2d(patches, image_size):
    """Reconstruct the image from all of its patches.
    Patches are assumed to overlap and the image is constructed by filling in
    the patches from left to right, top to bottom, averaging the overlapping
    regions.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    patches : array, shape = (n_patches, patch_height, patch_width) or
        (n_patches, patch_height, patch_width, n_channels)
        The complete set of patches. If the patches contain colour information,
        channels are indexed along the last dimension: RGB patches would
        have `n_channels=3`.
    image_size : tuple of ints (image_height, image_width) or
        (image_height, image_width, n_channels)
        the size of the image that will be reconstructed
    Returns
    -------
    image : array, shape = image_size
        the reconstructed image
    """
    i_h, i_w = image_size[:2]
    p_h, p_w = patches.shape[1:3]
    img = np.zeros(image_size)
    # compute the dimensions of the patches array
    n_h = i_h - p_h + 1
    n_w = i_w - p_w + 1
    for p, (i, j) in zip(patches, product(range(n_h), range(n_w))):
        img[i:i + p_h, j:j + p_w] += p

    for i in range(i_h):
        for j in range(i_w):
            # divide by the amount of overlap
            # XXX: is this the most efficient way? memory-wise yes, cpu wise?
            img[i, j] /= float(min(i + 1, p_h, i_h - i) *
                               min(j + 1, p_w, i_w - j))
    return img


def reconstruct_from_patches_3d(patches, volume_size):
    """Reconstruct the volume from all of its patches.
    Patches are assumed to overlap and the volume is constructed by filling in
    the patches from left to right, top to bottom, averaging the overlapping
    regions.
    Read more in the :ref:`User Guide <image_feature_extraction>`.
    Parameters
    ----------
    patches : array, shape = (n_patches, patch_x, patch_y, patch_z)
    volume_size : tuple of ints (volume_x, volume_y, volume_z)
        the size of the image that will be reconstructed
    Returns
    -------
    volume : array, shape = volume_size
        the reconstructed volume
    """
    v_x, v_y, v_z = volume_size[:3]
    p_x, p_y, p_z = patches.shape[1:4]
    vol = np.zeros(volume_size)
    # compute the dimensions of the patches array
    n_x = v_x - p_x + 1
    n_y = v_y - p_y + 1
    n_z = v_z - p_z + 1
    for p, (i, j, k) in zip(patches, product(range(n_x), range(n_y), range(n_z))):
        vol[i:i + p_x, j:j + p_y, k:k + p_z] += p

    for i in range(v_x):
        for j in range(v_y):
            for k in range(v_z):
                # divide by the amount of overlap
                # XXX: is this the most efficient way? memory-wise yes, cpu wise?
                vol[i, j, k] /= float(min(i + 1, p_x, v_x - i) *
                                      min(j + 1, p_y, v_y - j) *
                                      min(k + 1, p_z, v_z - k))
    return vol


# def resize(image, output_shape, order=1, mode='constant', cval=0, clip=True, preserve_range=False):
#     """Resize image to match a certain size.
#     Performs interpolation to up-size or down-size images. For down-sampling
#     N-dimensional images by applying the arithmetic sum or mean, see
#     `skimage.measure.local_sum` and `skimage.transform.downscale_local_mean`,
#     respectively.
#     Parameters
#     ----------
#     image : ndarray
#         Input image.
#     output_shape : tuple or ndarray
#         Size of the generated output image `(rows, cols[, dim])`. If `dim` is
#         not provided, the number of channels is preserved. In case the number
#         of input channels does not equal the number of output channels a
#         3-dimensional interpolation is applied.
#     Returns
#     -------
#     resized : ndarray
#         Resized version of the input.
#     Other parameters
#     ----------------
#     order : int, optional
#         The order of the spline interpolation, default is 1. The order has to
#         be in the range 0-5. See `skimage.transform.warp` for detail.
#     mode : {'constant', 'edge', 'symmetric', 'reflect', 'wrap'}, optional
#         Points outside the boundaries of the input are filled according
#         to the given mode.  Modes match the behaviour of `numpy.pad`.
#     cval : float, optional
#         Used in conjunction with mode 'constant', the value outside
#         the image boundaries.
#     clip : bool, optional
#         Whether to clip the output to the range of values of the input image.
#         This is enabled by default, since higher order interpolation may
#         produce values outside the given input range.
#     preserve_range : bool, optional
#         Whether to keep the original range of values. Otherwise, the input
#         image is converted according to the conventions of `img_as_float`.
#     Note
#     ----
#     Modes 'reflect' and 'symmetric' are similar, but differ in whether the edge
#     pixels are duplicated during the reflection.  As an example, if an array
#     has values [0, 1, 2] and was padded to the right by four values using
#     symmetric, the result would be [0, 1, 2, 2, 1, 0, 0], while for reflect it
#     would be [0, 1, 2, 1, 0, 1, 2].
#     Examples
#     --------
#     # >>> from skimage import data
#     # >>> from skimage.transform import resize
#     # >>> image = data.camera()
#     # >>> resize(image, (100, 100)).shape
#     (100, 100)
#     """
#
#     rows, cols = output_shape[0], output_shape[1]
#     orig_rows, orig_cols = image.shape[0], image.shape[1]
#
#     row_scale = float(orig_rows) / rows
#     col_scale = float(orig_cols) / cols
#
#     # 3-dimensional interpolation
#     if len(output_shape) == 3 and (image.ndim == 2
#                                    or output_shape[2] != image.shape[2]):
#         ndi_mode = _to_ndimage_mode(mode)
#         dim = output_shape[2]
#         if image.ndim == 2:
#             image = image[:, :, np.newaxis]
#         orig_dim = image.shape[2]
#         dim_scale = float(orig_dim) / dim
#
#         map_rows, map_cols, map_dims = np.mgrid[:rows, :cols, :dim]
#         map_rows = row_scale * (map_rows + 0.5) - 0.5
#         map_cols = col_scale * (map_cols + 0.5) - 0.5
#         map_dims = dim_scale * (map_dims + 0.5) - 0.5
#
#         coord_map = np.array([map_rows, map_cols, map_dims])
#
#         image = _convert_warp_input(image, preserve_range)
#
#         out = ndi.map_coordinates(image, coord_map, order=order,
#                                   mode=ndi_mode, cval=cval)
#
#         _clip_warp_output(image, out, order, mode, cval, clip)
#
#     else:  # 2-dimensional interpolation
#
#         if rows == 1 and cols == 1:
#             tform = AffineTransform(translation=(orig_cols / 2.0 - 0.5,
#                                                  orig_rows / 2.0 - 0.5))
#         else:
#             # 3 control points necessary to estimate exact AffineTransform
#             src_corners = np.array([[1, 1], [1, rows], [cols, rows]]) - 1
#             dst_corners = np.zeros(src_corners.shape, dtype=np.double)
#             # take into account that 0th pixel is at position (0.5, 0.5)
#             dst_corners[:, 0] = col_scale * (src_corners[:, 0] + 0.5) - 0.5
#             dst_corners[:, 1] = row_scale * (src_corners[:, 1] + 0.5) - 0.5
#
#             tform = AffineTransform()
#             tform.estimate(src_corners, dst_corners)
#
#         out = warp(image, tform, output_shape=output_shape, order=order,
#                    mode=mode, cval=cval, clip=clip,
#                    preserve_range=preserve_range)
#
#     return out


import os
import random
import nrrd
import matplotlib.pyplot as plt
from skimage.transform import rotate

def main(path_to_predictions, path_to_LR, path_to_HR, num_cases, path_to_save):
    predictions = os.listdir(path_to_predictions)
    LR, HR = [], []
    for arg_lr, arg_hr in zip(os.listdir(path_to_LR), os.listdir(path_to_HR)):
        if arg_lr.endswith('.nrrd'):
            LR.append(arg_lr)
        if arg_hr.endswith('.nrrd'):
            HR.append(arg_hr)
    
    # folder to save samples
    if not os.path.exists(path_to_save):
        os.makedirs(path_to_save)
    # select required number of cases
    random_case_idx = random.sample(range(0, len(predictions)-1), num_cases)
    for case in random_case_idx:
        print("Running case {}".format(predictions[case]))
        # read data
        prediction, pd_header = nrrd.read(os.path.join(path_to_predictions, predictions[case]))
        lr_GT, lr_header = nrrd.read(os.path.join(path_to_LR, LR[case]))
        hr_GT, hr_header = nrrd.read(os.path.join(path_to_HR, HR[case]))
        # get center slice
        ctr_slice = prediction.shape[2]//2
        # draw figure
        f, (ax0, ax1, ax2) = plt.subplots(nrows=1, ncols=3, figsize=(20, 10), dpi=200)
        ax0.imshow(rotate(lr_GT[:, :, ctr_slice], -90), cmap='gray')
        ax0.set_title('K1-D10')
        ax0.axis('off')
        ax1.imshow(rotate(prediction[:, :, ctr_slice], -90), cmap='gray')
        ax1.set_title('GAN-K2-D100')
        ax1.axis('off')
        ax2.imshow(rotate(hr_GT[:, :, ctr_slice], -90), cmap='gray')
        ax2.set_title('K2-D100')
        ax2.axis('off')
        # save images
        if path_to_save:
            file_name = predictions[case].split(".nrrd")[0] + ".png"
            file_save_path = os.path.join(path_to_save, file_name)
            f.savefig(file_save_path)
            f.clf()
    print("All files saved!")


if __name__ == "__main__":
    path_to_predictions = "/dingo/data/anil_dir/test_gan_dingo_k3_d10/gan_k3d10st1.0_TO_k2d100st1.0"
    path_to_LR = "/test_cases/nastaran_test_h5/k3_d10_st1.0"
    path_to_HR = "/test_cases/nastaran_test_h5/k2_d100_st1.0"
    save_path = "/workspace/NormGAN/visualize_samples/k3d10-k2d100"
    num_cases = 20
    main(path_to_predictions, path_to_LR, path_to_HR, num_cases, save_path)


import os
import cv2
import numpy as np
import pydicom as dicom
import h5py
import json


def _save_metadata(save_to, spacing, size, orientation, origin):
    header = {"PixelType": "short", "Dimension": "3", "Size": size, "Origin": origin, "Spacing": spacing, "Orientation": orientation, "MinPoint": "0 0 1", "Compression": "ZLib"}
    with open(save_to, 'w') as out_file:
        json.dump(header, out_file)

# check saving datatype of h5 file - overflow issue
def _save_h5(image, save_to):
    hf = h5py.File(save_to, 'w')
    hf.create_dataset('data', data=image, dtype='int16')
    hf.close()

def read_files(root, save_to):
    slices = [dicom.read_file(root + '/' + s) for s in os.listdir(root)]
    slices.sort(key=lambda x: x.InstanceNumber)
    image = np.stack([s.pixel_array for s in slices])
    # convert to int16
    image = image.astype(np.int16) 
    image[image == -2000] = 0
    # print('img min-max before linear processing:', image.min(), image.max())
    # convert to HU (all slices should have same RescaleIntercept and RescaleSlope(1))
    intercept = slices[0].RescaleIntercept
    slope = slices[0].RescaleSlope
    # doing a linear transform
    if slope != 1:
        image = slope * image.astype(np.float64)
        image = image.astype(np.int16)
    image += np.int16(intercept)
    
    image = np.clip(image, -1000, None)
    image += 1000
    # print('img min-max after cliping and addition:', image.min(), image.max())
    h5_file_name = os.path.join(save_to, '_'.join(root.split('/')[2:])+'.h5')
    h5_json_name = os.path.join(save_to, '_'.join(root.split('/')[2:])+'.json')
    _save_h5(image, h5_file_name)
    # information for metadata
    pixel_spacing = slices[0].PixelSpacing
    slice_thickness = slices[0].SliceThickness
    spacing = "{} {} {}".format(pixel_spacing[0], pixel_spacing[1], slice_thickness)
    size = [image.shape[1], image.shape[2], image.shape[0]]
    orientation = ' '.join(map(str, slices[0].ImageOrientationPatient)) # dummy orientation and origin information
    orientation = '-1 ' + orientation + ' -1'
    origin = "1 1 1"
    _save_metadata(h5_json_name, spacing, size, orientation, origin)

def main(root, save_to, include=None, exclude=None):
    if not os.path.exists(save_to):
        os.mkdir(save_to)
    for root, dirs, files in os.walk(root):
        if len(files) != 0:
            if include:
                if include in root:
                    if exclude:
                        if include+'_'+exclude not in root:
                            read_files(root, save_to)
                    else:
                        read_files(root, save_to)
            else:
                read_files(root, save_to)
    print('All files processed!')

"""
file = "/gan_data/UCLA/data/k2_d10_st1.0/af8923702f81618f9030387e9dd51589.h5"
with h5py.File(file, 'r') as file:
    IMG_THICKNESS, IMG_WIDTH, IMG_HEIGHT = file['data'].shape
    print(IMG_THICKNESS, IMG_WIDTH, IMG_HEIGHT)
    pixel_array = np.array(file['data'])
    print(pixel_array.dtype)
    print(pixel_array.min(), pixel_array.max())
"""

if __name__ == '__main__':
    main(root='/aapm_data', save_to='/workspace/NormGAN/aapm_3d_testset', include='full_1mm', exclude='sharp')


from __future__ import absolute_import

def create_model(opt):
    model = None
    print(opt.model)
    from .siam_model import *
    model = DistModel()
    model.initialize(opt, opt.batchSize, )
    print("model [%s] was created" % (model.name()))
    return model




import torch
import numpy as np
import os
from torch.utils.data import Dataset
import sys
import csv
from torch.utils.data import DataLoader
import pickle


class HscnnDataset(Dataset):
    def __init__(self, path_to_data, path_to_label_csv, num_low_level_tasks=3, transform=None):
        """
        Args:
            path_to_data (string): Path to images (.npy format supported)
            path_to_label_csv (string): CSV file with training/validation labels
            num_low_level_tasks (int, optional): number of low level tasks for the model
        """
        self.parsed_data = self._load_pickle(path_to_data)
        self.path_to_label_csv = path_to_label_csv
        self.transform = transform
        self.num_low_level_tasks = num_low_level_tasks
        try:
            with self._open_for_csv(self.path_to_label_csv) as file:
                # Dictionary with keys as file name and values as labels
                self.image_data = self._read_labels(csv.reader(file, delimiter=','))
        except ValueError as e:
            raise(ValueError('invalid CSV file: {}: {}'.format(self.path_to_data, e)), None)
        self.image_names = list(self.image_data.keys())

    def _load_pickle(self, path_to_data):
        """
        Reads a pickle file and return numpy array
        """
        with open(path_to_data, "rb") as f:
            file = pickle._Unpickler(f)
            file.encoding = 'latin1'
            file = file.load()
            return file

    def _open_for_csv(self, path):
        """
        Open a file with flags suitable for csv.reader. For python2, open with mode 'rb',
        for python3 this means 'r' with "universal newlines".
        """
        if sys.version_info[0] < 3:
            return open(path, 'rb')
        else:
            return open(path, 'r', newline='')

    def __len__(self):
        """
        :return: Number of samples in dataset
        """
        return len(self.image_names)

    def __getitem__(self, idx):
        """
        :param idx: index value of the image currently in memory
        :return: dictionary with image, label
        """
        img = self.load_image(idx, add_dim=True)
        label = self.load_labels(idx)
        if self.transform:
            img_transform = self.transform(img)
        sample = {'img': img, 'label': label}
        return sample

    def load_image(self, image_index, upper_bound=-500, lower_bound=-1000, add_dim=False):
        """
        :param image_index: index value of the image currently in memory
        :param upper_bound: upper bound for intensity values
        :param lower_bound: lower bound for intensity values
        :param add_dim: adds a channel dimension to the image
        :return: limits the image values between the two bounds and scales the image betwen 0-1
        """
        #path_to_img = os.path.join(self.path_to_data, self.image_names[image_index])
        img = self.parsed_data[image_index].astype(np.float32)
        img[img > upper_bound] = upper_bound
        img[img < lower_bound] = lower_bound
        #array = (img - lower_bound) / (upper_bound - lower_bound)
        array = (img - img.min()) / (img.max() - img.min())
        if add_dim:
            array = np.expand_dims(array, axis=3)
            array = array.transpose((3, 0, 1, 2))
        return torch.from_numpy(array)

    def load_labels(self, image_index):
        """
        :param image_index: index value of the image currently in memory
        :return: labels in tensor format
        """
        label_list = self.image_data[self.image_names[image_index]]
        if len(label_list) > 1:
            raise (ValueError('Multiple label found for image {}:'.format(self.image_names[image_index])), None)
        else:
            data = [int(arg) for arg in list(label_list[0].values())]
            tensor_lbl = torch.from_numpy(np.array(data))
        return tensor_lbl

    def _read_labels(self, csv_reader, header_available=True):
        """
        Returns a dictionary with index number as key and labels as values
        """
        if header_available:
            next(csv_reader)
        result = {}
        for line, col in enumerate(csv_reader):
            label_dict = {}
            try:
                img_file = str(line)
                if img_file not in result:
                    result[img_file] = []
                for i in range(len(col)):
                    label_dict['label'+str(i+1)] = col[i]
                result[img_file].append(label_dict)
            except ValueError:
                raise(ValueError(
                    'line {}: format should be \'label1,label2,label3,label4, img_file\' or \',,,,img_file\''.format(
                        line)), None)
        return result


from collections import namedtuple
import torch
from torchvision import models
from IPython import embed

class squeezenet(torch.nn.Module):
    def __init__(self, requires_grad=False, pretrained=True):
        super(squeezenet, self).__init__()
        pretrained_features = models.squeezenet1_1(pretrained=pretrained).features
        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()
        self.slice4 = torch.nn.Sequential()
        self.slice5 = torch.nn.Sequential()
        self.slice6 = torch.nn.Sequential()
        self.slice7 = torch.nn.Sequential()
        self.N_slices = 7
        for x in range(2):
            self.slice1.add_module(str(x), pretrained_features[x])
        for x in range(2,5):
            self.slice2.add_module(str(x), pretrained_features[x])
        for x in range(5, 8):
            self.slice3.add_module(str(x), pretrained_features[x])
        for x in range(8, 10):
            self.slice4.add_module(str(x), pretrained_features[x])
        for x in range(10, 11):
            self.slice5.add_module(str(x), pretrained_features[x])
        for x in range(11, 12):
            self.slice6.add_module(str(x), pretrained_features[x])
        for x in range(12, 13):
            self.slice7.add_module(str(x), pretrained_features[x])
        if not requires_grad:
            for param in self.parameters():
                param.requires_grad = False

    def forward(self, X):
        h = self.slice1(X)
        h_relu1 = h
        h = self.slice2(h)
        h_relu2 = h
        h = self.slice3(h)
        h_relu3 = h
        h = self.slice4(h)
        h_relu4 = h
        h = self.slice5(h)
        h_relu5 = h
        h = self.slice6(h)
        h_relu6 = h
        h = self.slice7(h)
        h_relu7 = h
        vgg_outputs = namedtuple("SqueezeOutputs", ['relu1','relu2','relu3','relu4','relu5','relu6','relu7'])
        out = vgg_outputs(h_relu1,h_relu2,h_relu3,h_relu4,h_relu5,h_relu6,h_relu7)

        return out


class alexnet(torch.nn.Module):
    def __init__(self, requires_grad=False, pretrained=True):
        super(alexnet, self).__init__()
        alexnet_pretrained_features = models.alexnet(pretrained=pretrained).features
        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()
        self.slice4 = torch.nn.Sequential()
        self.slice5 = torch.nn.Sequential()
        self.N_slices = 5
        for x in range(2):
            self.slice1.add_module(str(x), alexnet_pretrained_features[x])
        for x in range(2, 5):
            self.slice2.add_module(str(x), alexnet_pretrained_features[x])
        for x in range(5, 8):
            self.slice3.add_module(str(x), alexnet_pretrained_features[x])
        for x in range(8, 10):
            self.slice4.add_module(str(x), alexnet_pretrained_features[x])
        for x in range(10, 12):
            self.slice5.add_module(str(x), alexnet_pretrained_features[x])
        if not requires_grad:
            for param in self.parameters():
                param.requires_grad = False

    def forward(self, X):
        h = self.slice1(X)
        h_relu1 = h
        h = self.slice2(h)
        h_relu2 = h
        h = self.slice3(h)
        h_relu3 = h
        h = self.slice4(h)
        h_relu4 = h
        h = self.slice5(h)
        h_relu5 = h
        alexnet_outputs = namedtuple("AlexnetOutputs", ['relu1', 'relu2', 'relu3', 'relu4', 'relu5'])
        out = alexnet_outputs(h_relu1, h_relu2, h_relu3, h_relu4, h_relu5)
        return out


"""
Download vgg16 architecture, turn off 'requires_grad', divide layers into chunk of slices, 
and return output as namedtuple
"""
class vgg16(torch.nn.Module):
    def __init__(self, requires_grad=False, pretrained=True):
        super(vgg16, self).__init__()
        # download a pretrained vgg16  architecture (has 30 layers in total)
        vgg_pretrained_features = models.vgg16(pretrained=pretrained).features
        # create 5 empty sequential modules
        self.slice1 = torch.nn.Sequential()
        self.slice2 = torch.nn.Sequential()
        self.slice3 = torch.nn.Sequential()
        self.slice4 = torch.nn.Sequential()
        self.slice5 = torch.nn.Sequential()
        self.N_slices = 5

        # divide vgg16 layers into a chunck of slices, each slice must end with a ReLU layer
        # store the sliced layers into empty sequential modules defined above
        for x in range(4):
            self.slice1.add_module(str(x), vgg_pretrained_features[x])
        for x in range(4, 9):
            self.slice2.add_module(str(x), vgg_pretrained_features[x])
        for x in range(9, 16):
            self.slice3.add_module(str(x), vgg_pretrained_features[x])
        for x in range(16, 23):
            self.slice4.add_module(str(x), vgg_pretrained_features[x])
        for x in range(23, 30):
            self.slice5.add_module(str(x), vgg_pretrained_features[x])

        # print("slice1:", self.slice1)
        # print("-"*40)
        # print("slice2:", self.slice2)
        # print("-"*40)
        # print("slice3:", self.slice3)
        # print("-"*40)
        # print("slice4:", self.slice4)
        
        # 'requires_grad=False'; if not False is is True, so we set 'requires_grad' to be False
        if not requires_grad:
            for param in self.parameters():
                param.requires_grad = False

    def forward(self, X):
        # pass output through all the 5 slices/sequential layer modules 
        # also store output after passing through each slices
        h = self.slice1(X)
        h_relu1_2 = h
        h = self.slice2(h)
        h_relu2_2 = h
        h = self.slice3(h)
        h_relu3_3 = h
        h = self.slice4(h)
        h_relu4_3 = h
        h = self.slice5(h)
        h_relu5_3 = h
        # create a named tuple to store each of output of 5 layers
        vgg_outputs = namedtuple("VggOutputs", ['relu1_2', 'relu2_2', 'relu3_3', 'relu4_3', 'relu5_3'])
        out = vgg_outputs(h_relu1_2, h_relu2_2, h_relu3_3, h_relu4_3, h_relu5_3)
        return out # return 'namedtuple'


class resnet(torch.nn.Module):
    def __init__(self, requires_grad=False, pretrained=True, num=18):
        super(resnet, self).__init__()
        if(num==18):
            self.net = models.resnet18(pretrained=pretrained)
        elif(num==34):
            self.net = models.resnet34(pretrained=pretrained)
        elif(num==50):
            self.net = models.resnet50(pretrained=pretrained)
        elif(num==101):
            self.net = models.resnet101(pretrained=pretrained)
        elif(num==152):
            self.net = models.resnet152(pretrained=pretrained)
        self.N_slices = 5

        self.conv1 = self.net.conv1
        self.bn1 = self.net.bn1
        self.relu = self.net.relu
        self.maxpool = self.net.maxpool
        self.layer1 = self.net.layer1
        self.layer2 = self.net.layer2
        self.layer3 = self.net.layer3
        self.layer4 = self.net.layer4

    def forward(self, X):
        h = self.conv1(X)
        h = self.bn1(h)
        h = self.relu(h)
        h_relu1 = h
        h = self.maxpool(h)
        h = self.layer1(h)
        h_conv2 = h
        h = self.layer2(h)
        h_conv3 = h
        h = self.layer3(h)
        h_conv4 = h
        h = self.layer4(h)
        h_conv5 = h

        outputs = namedtuple("Outputs", ['relu1','conv2','conv3','conv4','conv5'])
        out = outputs(h_relu1, h_conv2, h_conv3, h_conv4, h_conv5)

        return out



import numpy as np

# deafult nms in python for checking
def py_nms(dets, thresh):
    z = dets[:, 1]
    y = dets[:, 2]
    x = dets[:, 3]
    d = dets[:, 4]
    h = dets[:, 5]
    w = dets[:, 6]
    scores = dets[:, 0]

    areas = d * h * w
    order = scores.argsort()[::-1]
    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)

        xx0 = np.maximum(x[i] - w[i] / 2, x[order[1:]] - w[order[1:]] / 2)
        yy0 = np.maximum(y[i] - h[i] / 2, y[order[1:]] - h[order[1:]] / 2)
        zz0 = np.maximum(z[i] - d[i] / 2, z[order[1:]] - d[order[1:]] / 2)
        xx1 = np.minimum(x[i] + w[i] / 2, x[order[1:]] + w[order[1:]] / 2)
        yy1 = np.minimum(y[i] + h[i] / 2, y[order[1:]] + h[order[1:]] / 2)
        zz1 = np.minimum(z[i] + d[i] / 2, z[order[1:]] + d[order[1:]] / 2)

        inter_w = np.maximum(0.0, xx1 - xx0)
        inter_h = np.maximum(0.0, yy1 - yy0)
        inter_d = np.maximum(0.0, zz1 - zz0)
        intersect = inter_w * inter_h * inter_d
        overlap = intersect / (areas[i] + areas[order[1:]] - intersect)

        inds = np.where(overlap <= thresh)[0]
        order = order[inds + 1]

    return dets[keep], keep


from setuptools import setup, find_packages
import codecs
import os

VERSION = '0.0.5'
DESCRIPTION = 'A lung analysis pipeline'
LONG_DESCRIPTION = 'A lung analysis pipeline for normalization, detection, segmentation, and feature extraction'

# Setting up
setup(
    name="lung_analysis_pipeline",
    version=VERSION,
    author="Olivia Zhang",
    author_email="<olivia.zhang@ucla.edu>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[
        'numpy',
        'torch'
    ], 
    # include_package_data=True,
)


from setting import parse_opts 
from datasets.brains18 import BrainS18Dataset
from model import generate_model
import torch
import numpy as np
from torch.utils.data import DataLoader
import torch.nn.functional as F
from scipy import ndimage
import nibabel as nib
import sys
import os
from utils.file_process import load_lines
import numpy as np


def seg_eval(pred, label, clss):
    """
    calculate the dice between prediction and ground truth
    input:
        pred: predicted mask
        label: groud truth
        clss: eg. [0, 1] for binary class
    """
    Ncls = len(clss)
    dices = np.zeros(Ncls)
    [depth, height, width] = pred.shape
    for idx, cls in enumerate(clss):
        # binary map
        pred_cls = np.zeros([depth, height, width])
        pred_cls[np.where(pred == cls)] = 1
        label_cls = np.zeros([depth, height, width])
        label_cls[np.where(label == cls)] = 1

        # cal the inter & conv
        s = pred_cls + label_cls
        inter = len(np.where(s >= 2)[0])
        conv = len(np.where(s >= 1)[0]) + inter
        try:
            dice = 2.0 * inter / conv
        except:
            print("conv is zeros when dice = 2.0 * inter / conv")
            dice = -1

        dices[idx] = dice

    return dices


def test(data_loader, model, img_names, sets):
    masks = []
    model.eval() # for testing 
    for batch_id, batch_data in enumerate(data_loader):
        # forward
        volume = batch_data
        if not sets.no_cuda:
            volume = volume.cuda()
        with torch.no_grad():
            probs = model(volume)
            probs = F.softmax(probs, dim=1)

        # resize mask to original size
        [batchsize, _, mask_d, mask_h, mask_w] = probs.shape
        data = nib.load(os.path.join(sets.data_root, img_names[batch_id]))
        data = data.get_data()
        [depth, height, width] = data.shape
        mask = probs[0]
        scale = [1, depth*1.0/mask_d, height*1.0/mask_h, width*1.0/mask_w]
        mask = ndimage.interpolation.zoom(mask, scale, order=1)
        mask = np.argmax(mask, axis=0)
        masks.append(mask)
    return masks


if __name__ == '__main__':
    # settting
    sets = parse_opts()
    sets.target_type = "normal"
    sets.phase = 'test'

    # getting model
    checkpoint = torch.load(sets.resume_path)
    net, _ = generate_model(sets)
    net.load_state_dict(checkpoint['state_dict'])

    # data tensor
    testing_data =BrainS18Dataset(sets.data_root, sets.img_list, sets)
    data_loader = DataLoader(testing_data, batch_size=1, shuffle=False, num_workers=1, pin_memory=False)

    # testing
    img_names = [info.split(" ")[0] for info in load_lines(sets.img_list)]
    masks = test(data_loader, net, img_names, sets)
    
    # evaluation: calculate dice 
    label_names = [info.split(" ")[1] for info in load_lines(sets.img_list)]
    Nimg = len(label_names)
    dices = np.zeros([Nimg, sets.n_seg_classes])
    for idx in range(Nimg):
        label = nib.load(os.path.join(sets.data_root, label_names[idx]))
        label = label.get_data()
        dices[idx, :] = seg_eval(masks[idx], label, range(sets.n_seg_classes))
    
    # print result
    for idx in range(1, sets.n_seg_classes):
        mean_dice_per_task = np.mean(dices[:, idx])
        print('mean dice for class-{} is {}'.format(idx, mean_dice_per_task))   



import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
# from torchvision.models.feature_extraction import create_feature_extractor
from torchvision import transforms


class VGGContentLoss(nn.Module):
    """
    Constructs a content loss function based on the VGG19 network.
    Using high-level feature mapping layers from the latter layers will focus more on the texture content of the image.
    Paper reference list:
        -`Photo-Realistic Single Image Super-Resolution Using a Generative Adversarial Network <https://arxiv.org/pdf/1609.04802.pdf>` paper.
        -`ESRGAN: Enhanced Super-Resolution Generative Adversarial Networks                    <https://arxiv.org/pdf/1809.00219.pdf>` paper.
        -`Perceptual Extreme Super Resolution Network with Receptive Field Block               <https://arxiv.org/pdf/2005.12597.pdf>` paper.
     """
    def __init__(self, feature_model_extractor_node: str,
                 feature_model_normalize_mean: list,
                 feature_model_normalize_std: list):
        super(VGGContentLoss, self).__init__()
        self.feature_model_extractor_node = feature_model_extractor_node
        model = models.vgg19(True)
        # Extract the thirty-fifth layer output in the VGG19 model as the content loss.
        self.feature_extractor = create_feature_extractor(model, [feature_model_extractor_node])
        # set to validation mode
        self.feature_extractor.eval()
        self.normalize = transforms.Normalize(feature_model_normalize_mean, feature_model_normalize_std)
        # Freeze model parameters
        for model_parameters in self.feature_extractor.parameters():
            model_parameters.requires_grad = False

    def forward(self, sr_tensor: torch.Tensor, hr_tensor: torch.Tensor) -> torch.Tensor:
        # print('Inside `ContentLoss` forward function')
        # Standardized operations
        # print('BeforeNorm - SR min-max:', sr_tensor.min(), sr_tensor.max())
        # print('BeforeNorm - HR min-max:', hr_tensor.min(), hr_tensor.max())
        sr_tensor = self.normalize(sr_tensor) # 0 to 1 (before norm) | -2 to 2 (after norm)
        hr_tensor = self.normalize(hr_tensor)

        # print('AfterNorm - SR min-max:', sr_tensor.min(), sr_tensor.max())
        # print('AfterNorm - HR min-max:', hr_tensor.min(), hr_tensor.max())

        sr_feature = self.feature_extractor(sr_tensor)[self.feature_model_extractor_node]
        hr_feature = self.feature_extractor(hr_tensor)[self.feature_model_extractor_node]

        # Find the feature map difference between the two images
        content_loss = F.l1_loss(sr_feature, hr_feature)
        return content_loss











"""
Perceptual Network - Assume input range is [0, 1]
"""
"""
class VGGFeatureExtractor(nn.Module):
    def __init__(self,
                 feature_layer=34,
                 use_bn=False,
                 use_input_norm=True, # normalizes input tensor before feeding to model
                 device=torch.device('cpu')):
        super(VGGFeatureExtractor, self).__init__()
        # if 'use_bn'=True; initialize vgg19 with batchnorm
        if use_bn:
            model = torchvision.models.vgg19_bn(pretrained=True)
        else:
            # if 'use_bn'=True; initialize vgg19 without batchnorm
            model = torchvision.models.vgg19(pretrained=True)
        
        self.use_input_norm = use_input_norm
        if self.use_input_norm:
            # define 'mean' and 'std' variable (for normalization) of size similar to input data
            mean = torch.Tensor([0.485, 0.456, 0.406]).view(1, 3, 1, 1).to(device)
            # [0.485-1, 0.456-1, 0.406-1] if input in range [-1,1]
            std = torch.Tensor([0.229, 0.224, 0.225]).view(1, 3, 1, 1).to(device)
            # [0.229*2, 0.224*2, 0.225*2] if input in range [-1,1]

            # creates 'self.mean' and 'self.std' attribute with value defined by above 'mean' and 'std'
            self.register_buffer('mean', mean)
            self.register_buffer('std', std)
        
        # unpack modules in 'nn.Sequential' upto the defined 'feature_layer' before activation
        self.features = nn.Sequential(*list(model.features.children())[:(feature_layer + 1)])
        
        # No need to Back Propogate the parameters, so we set 'require_grad' to False
        for k, v in self.features.named_parameters():
            v.requires_grad = False

    def forward(self, x):
        # if 'self.use_input_norm' is True, normalize data by 'self.mean' and 'self.std'
        if self.use_input_norm:
            x = (x - self.mean) / self.std
        # get output and return
        output = self.features(x)
        return output
"""

class MINCNet(nn.Module):
    def __init__(self):
        super(MINCNet, self).__init__()
        self.ReLU = nn.ReLU(True)
        self.conv11 = nn.Conv2d(3, 64, 3, 1, 1)
        self.conv12 = nn.Conv2d(64, 64, 3, 1, 1)
        self.maxpool1 = nn.MaxPool2d(2, stride=2, padding=0, ceil_mode=True)
        self.conv21 = nn.Conv2d(64, 128, 3, 1, 1)
        self.conv22 = nn.Conv2d(128, 128, 3, 1, 1)
        self.maxpool2 = nn.MaxPool2d(2, stride=2, padding=0, ceil_mode=True)
        self.conv31 = nn.Conv2d(128, 256, 3, 1, 1)
        self.conv32 = nn.Conv2d(256, 256, 3, 1, 1)
        self.conv33 = nn.Conv2d(256, 256, 3, 1, 1)
        self.maxpool3 = nn.MaxPool2d(2, stride=2, padding=0, ceil_mode=True)
        self.conv41 = nn.Conv2d(256, 512, 3, 1, 1)
        self.conv42 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv43 = nn.Conv2d(512, 512, 3, 1, 1)
        self.maxpool4 = nn.MaxPool2d(2, stride=2, padding=0, ceil_mode=True)
        self.conv51 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv52 = nn.Conv2d(512, 512, 3, 1, 1)
        self.conv53 = nn.Conv2d(512, 512, 3, 1, 1)

    def forward(self, x):
        out = self.ReLU(self.conv11(x))
        out = self.ReLU(self.conv12(out))
        out = self.maxpool1(out)
        out = self.ReLU(self.conv21(out))
        out = self.ReLU(self.conv22(out))
        out = self.maxpool2(out)
        out = self.ReLU(self.conv31(out))
        out = self.ReLU(self.conv32(out))
        out = self.ReLU(self.conv33(out))
        out = self.maxpool3(out)
        out = self.ReLU(self.conv41(out))
        out = self.ReLU(self.conv42(out))
        out = self.ReLU(self.conv43(out))
        out = self.maxpool4(out)
        out = self.ReLU(self.conv51(out))
        out = self.ReLU(self.conv52(out))
        out = self.conv53(out)
        return out


# Assume input range is [0, 1]
class MINCFeatureExtractor(nn.Module):
    def __init__(self, feature_layer=34, use_bn=False, use_input_norm=True, \
                device=torch.device('cpu')):
        super(MINCFeatureExtractor, self).__init__()

        self.features = MINCNet()
        self.features.load_state_dict(
            torch.load('../experiments/pretrained_models/VGG16minc_53.pth'), strict=True)
        self.features.eval()
        # No need to BP to variable
        for k, v in self.features.named_parameters():
            v.requires_grad = False

    def forward(self, x):
        output = self.features(x)
        return output



import argparse
import os
from IPython import embed
from util import util
import models.dist_model as dm

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
# parser.add_argument('--dir0', type=str, default='./imgs/ex_dir0')
# parser.add_argument('--dir1', type=str, default='./imgs/ex_dir1')
parser.add_argument('--dir0', type=str, default='/home/leihaowei/datasets/UCLA/png/val/k2_d100_st1.0')
parser.add_argument('--dir1', type=str, default='/home/leihaowei/MII/myrepo/SRGAN/results/SRGAN_denoise_003/val_num15_d10_d100_k2_st1')
parser.add_argument('--out', type=str, default='./imgs/SRGAN_denoise_003.txt')
parser.add_argument('--use_gpu', action='store_true', help='turn on flag to use GPU')
opt = parser.parse_args()

## Initializing the model
model = dm.DistModel()
model.initialize(model='net-lin',net='vgg',use_gpu=opt.use_gpu)

# crawl directories
f = open(opt.out,'w')

# files = os.listdir(opt.dir0)
files = []
for root, dirpath, fnames in sorted(os.walk(opt.dir0)):
    for fname in sorted(fnames):
        img_path = os.path.join(root, fname)
        # split path and combine
        img_path=os.path.join(os.path.split(os.path.split(img_path)[0])[-1], \
        	os.path.split(img_path)[-1])
        files.append(img_path)

avg_dist = 0.0
for file in files:
	if(os.path.exists(os.path.join(opt.dir1,file))):
		# Load images
		img0 = util.im2tensor(util.load_image(os.path.join(opt.dir0,file)), factor=1500./2.) # RGB image from [-1,1]
		img1 = util.im2tensor(util.load_image(os.path.join(opt.dir1,file)), factor=1500./2.)

		# Compute distance
		dist01 = model.forward(img0,img1)
		avg_dist += dist01
		print('%s: %.6f'%(file,dist01))
		f.writelines('%s: %.6f\n'%(file,dist01))
avg_dist = avg_dist / len(files)
f.writelines('average LPIPS: %.6f\n'%(avg_dist))
f.close()



import numpy as np


def box_transform(windows, targets, weight):
    """
    Calculate regression terms, dz, dy, dx, dd, dh, dw
    # windows should equal to # targets
    windows: [num_window, z, y, x, D, H, W]
    targets: [num_target, z, y, x, D, H, W]
    """
    wz, wy, wx, wd, wh, ww = weight
    bz, by, bx = windows[:, 0], windows[:, 1], windows[:, 2]
    bd, bh, bw = windows[:, 3], windows[:, 4], windows[:, 5]
    tz, ty, tx = targets[:, 0], targets[:, 1], targets[:, 2]
    td, th, tw = targets[:, 3], targets[:, 4], targets[:, 5]

    dz = wz * (tz - bz) / bd
    dy = wy * (ty - by) / bh
    dx = wx * (tx - bx) / bw
    dd = wd * np.log(td / bd)
    dh = wh * np.log(th / bh)
    dw = ww * np.log(tw / bw)

    deltas = np.vstack((dz, dy, dx, dd, dh, dw)).transpose()
    return deltas


def box_transform_inv(windows, deltas, weight):
    """
    Apply regression terms to predicted bboxes
    windows: [num_window, z, y, x, D, H, W]
    targets: [num_target, z, y, x, D, H, W]
    """
    num   = len(windows)
    wz, wy, wx, wd, wh, ww = weight
    predictions = np.zeros((num, 6), dtype=np.float32)

    bz, by, bx = windows[:, 0], windows[:, 1], windows[:, 2]
    bd, bh, bw = windows[:, 3], windows[:, 4], windows[:, 5]
    bz = bz[:, np.newaxis]
    by = by[:, np.newaxis]
    bx = bx[:, np.newaxis]
    bd = bd[:, np.newaxis]
    bh = bh[:, np.newaxis]
    bw = bw[:, np.newaxis]

    dz = deltas[:, 0::6] / wz
    dy = deltas[:, 1::6] / wy
    dx = deltas[:, 2::6] / wx
    dd = deltas[:, 3::6] / wd
    dh = deltas[:, 4::6] / wh
    dw = deltas[:, 5::6] / ww

    z = dz * bd + bz
    y = dy * bh + by
    x = dx * bw + bx
    
    d = np.exp(dd) * bd
    h = np.exp(dh) * bh
    w = np.exp(dw) * bw

    predictions[:, 0::6] = z
    predictions[:, 1::6] = y
    predictions[:, 2::6] = x 
    predictions[:, 3::6] = d
    predictions[:, 4::6] = h
    predictions[:, 5::6] = w

    return predictions


def clip_boxes(boxes, img_size):
    """
    clip boxes outside the image, all box follows [p, z, y, x, d, h, w]
    """
    depth, height, width = img_size
    boxes[:, 0] = np.clip(boxes[:, 0], 0, depth - 1)
    boxes[:, 3] = np.clip(boxes[:, 3], 0, depth - 1)
    boxes[:, 1] = np.clip(boxes[:, 1], 0, height - 1)
    boxes[:, 4] = np.clip(boxes[:, 4], 0, height - 1)
    boxes[:, 2] = np.clip(boxes[:, 2], 0, width  - 1)
    boxes[:, 5] = np.clip(boxes[:, 5], 0, width - 1)

    return boxes


import os, time, glob
import math
from datetime import datetime
import pydicom
import numpy as np
import cv2
import random
import torch
import logging
import nrrd
import json
# from pydicom.dataset import Dataset, FileDataset
import SimpleITK as sitk
from ..PerceptualSimilarity.models import dist_model as dm
import re



"""
Get current timestamp in ymd-HMS format. Used when creataing a directory that already exists
"""
def get_timestamp():
    return datetime.now().strftime('%y%m%d-%H%M%S')

"""
Create a folder if it doesn't exist
"""
def mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

"""
Make mupltiple directories, if it is an instance of string
"""
def mkdirs(paths):
    if isinstance(paths, str):
        mkdir(paths)
    else:
        for path in paths:
            mkdir(path)

"""
Create a folder if it doesn't exits, otherwise create a renamed version if it already exists
"""
def mkdir_and_rename(path):
    if os.path.exists(path):
        new_name = path + '_archived_' + get_timestamp()
        print('Path already exists. Rename it to [{:s}]'.format(new_name))
        logger = logging.getLogger('base')
        logger.info('Path already exists. Rename it to [{:s}]'.format(new_name))
        os.rename(path, new_name)
    os.makedirs(path)


"""
Set seed for numpy and torch arrays
"""
def set_random_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


"""
Create two logger named 'base' and 'val', with file name specified in 'phase' variable at the 
path specified in 'root' variable.
If 'screen=True', message is printed to screen
"""
def setup_logger(logger_name, root, phase, level=logging.INFO, screen=False):
    '''set up logger'''
    l = logging.getLogger(logger_name)
    formatter = logging.Formatter(
        '%(asctime)s.%(msecs)03d - %(levelname)s: %(message)s', datefmt='%y-%m-%d %H:%M:%S')
    log_file = os.path.join(root, phase + '_{}.log'.format(get_timestamp()))
    fh = logging.FileHandler(log_file, mode='w')
    fh.setFormatter(formatter)
    l.setLevel(level)
    l.addHandler(fh)
    if screen:
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        l.addHandler(sh)

"""
Prints evaluaation metrics on screen
"""
def print_metrics(logger, metric_name, results):
    means = {}
    for view in list(results[list(results.keys())[0]].keys()):
        count = 0
        sum = 0.
        for k, v in results.items():
            count += 1
            sum += v[view]
            logger.info('{} for {} in {} view : {:.6f}'.format(metric_name, k, view, v[view]))
        means[view] = sum/count
        logger.info('Average {} in {} view : {:.6f}'.format(metric_name, view, means[view]))
    return means


####################
# image operations
####################
def tensor2img(tensor, out_type=np.uint8, min_max=(0, 1), intercept=-1024):
    '''
    Converts a torch Tensor into an image Numpy array
    Input: 3D(D,H,W)
    Output: 3D(H,W,D), np.uint8 (default), [0,1500] uint16, [-1000, 500] int16
    intercept is used when out_type is set to int16
    '''
    tensor = tensor.cpu().clamp_(*min_max)  # clamp
    tensor = (tensor - min_max[0]) / (min_max[1] - min_max[0])  # to range [0,1]
    img_np = tensor.numpy()
    if out_type == np.uint8:
        img_np = (img_np * 255.0).round()
        # Important. Unlike matlab, numpy.uint8() WILL NOT round by default.
    if out_type == np.uint16:
        img_np = (img_np * 1500.0).round()
    if out_type == np.int16:
        img_np = (img_np * 1500.0).round() + intercept 
    return img_np.astype(out_type)


def img2tensor(image, cent=1., factor=255./2.):
    '''
    :Input 3D(H,W,D):
    :Output (1,D,H,W) [-1,1]
    '''
    return torch.Tensor((image / factor - cent)[:, :, :, np.newaxis].transpose((3, 2, 0, 1)))


"""
Converts incoming h5 data input (uint16 - D,H,W) to torch tensor float in a given range range [0, 1500].
For dataloader use, inplace operation
"""
class ImgToTensor(object):
    def __call__(self, sample, clip=False, raw_data_range=1500.):
        if clip:
            img = torch.from_numpy(np.clip(np.array(sample, np.float32, copy=False), 0, raw_data_range))
        else:
            img = torch.from_numpy(np.array(sample, np.float32, copy=False))
        return img.float().div_(raw_data_range)  # range is from 0 to 1500 for png


"""
Read json configuration file for each image and returns the spacing, orientation and origin of the image
"""
def read_config(config_path):
    config_path = os.path.join(config_path) 
    with open(config_path, 'r') as f:
        config = json.load(f)
    meta_data = {}
    meta_data['Spacing'] = [float(i) for i in config['Spacing'].split()]
    meta_data['Orientation'] = [float(i) for i in config['Orientation'].split()] 
    meta_data['Origin'] = [float(i) for i in config['Origin'].split()]
    return meta_data

def save_vol(opt, spacings, results_dir, patient_id, volume):
    # nrrd accepts in [W,H,D] format
    # [D,H,W]==>[W,H,D]
    # no need to transpose, just add index_order='C'
    # volume = np.transpose(volume, (2, 1, 0))
    if len(spacings) != 0:
        copy_spacings = spacings.copy()
        copy_spacings[2] /= opt['scale']
    else:
        copy_spacings = []
        
    # uid = os.path.splitext(os.path.basename(volume_path))[0]
    # if opt['datasets']['val']['data_type'] == 'h5':
    #     # get spacing from json file
    #     config_path = os.path.join(opt['datasets']['val']['dataroot_LR'], uid + '.json')
    #     meta_data = read_config(config_path)
    #     spacings = meta_data['Spacing']
    #     orientations = meta_data['Orientation']
    #     # nrrd requires nested list for orientations
    #     # https://readthedocs.org/projects/pynrrd/downloads/pdf/latest/
    #     # http://teem.sourceforge.net/nrrd/format.html#spacedirections
    #     orientations = [orientations[i:i+3] for i in range(0, len(orientations), 3)]
    #     origin = meta_data['Origin']         
    #     spacings[2] /= opt['scale'] 
    # elif opt['datasets']['val']['data_type'] == 'dicom':
    #     # TODO: implement save according to input dicom
    #     spacings[2] /= opt['scale'] 
    # else: 
    #     raise NotImplementedError('supported output format: nrrd or dicom')
    volume_path = os.path.join(results_dir, patient_id+'.nrrd')
    if opt['datasets']['val']['data_type'] == 'nrrd':
        header = {'units': ['mm', 'mm', 'mm'], 'spacings': copy_spacings}
        nrrd.write(volume_path, volume, header, index_order='C')
        return
    header = {'units': ['mm', 'mm', 'mm'], 'spacings': copy_spacings} 
    nrrd.write(volume_path, volume, header, index_order='C')

def save_vol_nested(opt, spacings, results_dir, input_info, volume):
    patient_id = input_info['patient_id'][0]
    time = input_info['time'][0]
    name = input_info['name'][0]
    # create nested directories if not exist 
    patient_dir = os.path.join(results_dir, patient_id)
    mkdir(patient_dir)
    time_dir = os.path.join(patient_dir, time)
    mkdir(time_dir)
    name_dir = os.path.join(time_dir, name)
    mkdir(name_dir) 
    mkdir(os.path.join(name_dir, 'NRRD'))
    volume_path = os.path.join(name_dir, 'NRRD', '1.nrrd')
    # nrrd accepts in [W,H,D] format
    # [D,H,W]==>[W,H,D]
    # no need to transpose, just add index_order='C'
    # volume = np.transpose(volume, (2, 1, 0))
    if len(spacings) != 0:
        copy_spacings = spacings.copy()
        copy_spacings[2] /= opt['scale']
    else:
        copy_spacings = []

    header = {'units': ['mm', 'mm', 'mm'], 'spacings': copy_spacings}
    nrrd.write(volume_path, volume, header, index_order='C')


def save_dicoms(opt, spacings, volume_path, volume):
    mkdir(volume_path)
    copy_spacings = spacings.copy()
    copy_spacings[2] /= opt['scale'] 
    # if opt['datasets']['val']['data_type'] == 'h5':     
    #     config_path =  os.path.join(opt['datasets']['val']['dataroot_LR'], uid + '.json')
    #     meta_data = read_config(config_path)
    #     spacings = meta_data['Spacing']
    #     spacings[2] /= opt['scale'] 
    #     orientations = meta_data['Orientation']
    #     orientations = [orientations[i:i+3] for i in range(0, len(orientations), 3)]
    #     origin = meta_data['Origin']  
    # elif opt['datasets']['val']['data_type'] == 'dicom':
    #     spacings[2] /= opt['scale'] 
    # else: 
    #     raise NotImplementedError('supported output format: nrrd or dicom')
    write_dicom(copy_spacings, volume, volume_path)            
    print("written as dcm completed")
    # for i, pixel_array in enumerate(vol): 
    #     write_dicom(i+1, opt, meta_data, pixel_array, volume_path)


def write_dicom(spacings, new_arr, path):
    new_img = sitk.GetImageFromArray(new_arr)
    new_img.SetSpacing(spacings)
    
    writer = sitk.ImageFileWriter()
    # Use the study/series/frame of reference information given in the meta-data
    # dictionary and not the automatically generated information from the file IO
    writer.KeepOriginalImageUIDOn()

    modification_time = time.strftime("%H%M%S")
    modification_date = time.strftime("%Y%m%d")

    # Copy some of the tags and add the relevant tags indicating the change.
    # For the series instance UID (0020|000e), each of the components is a number, cannot start
    # with zero, and separated by a '.' We create a unique series ID using the date and time.
    # tags of interest:
    direction = new_img.GetDirection()
    series_tag_values = [("0008|0031",modification_time), # Series Time
                    ("0008|0021",modification_date), # Series Date
                    ("0008|0008","DERIVED\\SECONDARY"), # Image Type
                    ("0020|000e", "1.2.826.0.1.3680043.2.1125."+modification_date+".1"+modification_time), # Series Instance UID
                    ("0020|0037", '\\'.join(map(str, (direction[0], direction[3], direction[6],# Image Orientation (Patient)
                                                    direction[1],direction[4],direction[7])))),
                    ("0008|103e", "Normalized image"), # series description
                    ("0020|000d", "1.2.826.0.1.3680043.2.1125."+modification_date+".1"+modification_time), # study instance UID
                    ("0008|0020", modification_date), # study date
                    ("0008|0030", modification_time),  # study time
                    # Setting the type to CT preserves the slice location.
                    # set the type to CT so the thickness is carried over
                    ("0008|0060", "CT"),
                    ("0028|1050", "-600"), # WindowCenter
                    ("0028|1051", "1500"), # WindowWidth
                    ("0028|1054", "HU"), # RescaleType
                    ("0018|0050", str(spacings[2]))  # slice thickness
                    ]  

    # Write slices to output directory
    list(map(lambda i: writeSlices(writer, i, series_tag_values, new_img, path), range(new_img.GetDepth())))

def writeSlices(writer, index, series_tag_values, new_img, path):
    # DicomSeriesFromArray
    #https://simpleitk.readthedocs.io/en/next/Examples/DicomSeriesFromArray/Documentation.html
    image_slice = new_img[:,:,index]

    # Tags shared by the series.
    list(map(lambda tag_value: image_slice.SetMetaData(tag_value[0], tag_value[1]), series_tag_values))

    # Slice specific tags.
    image_slice.SetMetaData("0008|0012", time.strftime("%Y%m%d")) # Instance Creation Date
    image_slice.SetMetaData("0008|0013", time.strftime("%H%M%S")) # Instance Creation Time

    # (0020, 0032) image position patient determines the 3D spacing between slices.
    pos = [str(i) for i in new_img.TransformIndexToPhysicalPoint((0,0,index))]
    image_slice.SetMetaData("0020|0032", '\\'.join(pos)) # Image Position (Patient)
    image_slice.SetMetaData("0020|1041", pos[2]) # Slice location
    image_slice.SetMetaData("0020|0013", str(index+1)) # Instance Number

    # No need to set intercept/slope in ITK
    # RescaleIntercept 
    # image_slice.SetMetaData("0028|1052", "-1000")
    # RescaleSlope
    # image_slice.SetMetaData("0028|1053", "1")
    
    # Write to the output directory and add the extension dcm, to force writing in DICOM format.
    writer.SetFileName(os.path.join(path, '{0:0=3d}.dcm'.format(index+1)))
    writer.Execute(image_slice)

# need some work to get it working using pydicom
def write_pydicom(index, opt, meta_data, pixel_array, path):
    fullpath = os.path.join(path, '{0:0=3d}.dcm'.format(index))
    pixel_array = pixel_array.astype(np.int16)

    file_meta = Dataset()
    file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
    file_meta.MediaStorageSOPInstanceUID = "1.2.840.113654.2.55.102869246940549636091903990697209280630"
    file_meta.ImplementationClassUID = '1.2.40.0.13.1.1.1'
    file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1' 

    ds = FileDataset(fullpath, {},file_meta = file_meta, preamble=b"\0"*128)

    # ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
    # ds.SOPInstanceUID = "1.2.840.113654.2.55.102869246940549636091903990697209280630"
    ds.SeriesDescription = 'Normalized images'
    ds.SeriesInstanceUID = '1.2.840.113654.2.55.102869246940549636091903990697209280630'
    # ds.StudyInstanceUID =  '1.2.840.113654.2.55.102869246940549636091903990697209280630'
    # ds.StudyID = ''
    # ds.ContentDate = str(datetime.date.today()).replace('-','')
    # ds.ContentTime = str(time.time()) #milliseconds since the epoch
    ds.Modality = "CT"
    ds.SamplesPerPixel = 1
    ds.PhotometricInterpretation = "MONOCHROME2"
    ds.PixelRepresentation = 0
    ds.HighBit = 11
    ds.BitsStored = 12
    ds.BitsAllocated = 16
    ds.Columns, ds.Rows = pixel_array.shape
    ds.SeriesNumber = 1
    ds.AcqusitionNumber = 1
    
    ds.RescaleSlope = "1"
    ds.RescaleIntercept = "-1000"
    ds.RescaleType  = "HU"
    ds.WindowCenter = "-600"
    ds.WindowWidth = "1500"    
    ds.InstanceNumber = index


    spacings = meta_data['Spacing']
    spacings[2] /= opt['scale'] 
    # orientations = meta_data['Orientation']
    # origin = meta_data['Origin']  

    # ds.ImagePositionPatient = origin       
    # ds.SliceLocation = origin[2] - (index - 1) * spacings[2]   
    # ds.ImageOrientationPatient = orientations[:6]
    ds.SliceThickness = spacings[2] + 1e-10 
    ds.PixelSpacing = [spacings[0], spacings[1]]
    ds.PixelData = pixel_array.tobytes()
    ds.save_as(fullpath)
    

####################
# metric
####################
def calculate_psnr(img1, img2):
    # img1 and img2 have range [0, 1]
    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    mse = np.mean((img1 - img2)**2)
    if mse == 0:
        return float('inf')
    return 20 * math.log10(255.0 / math.sqrt(mse))


def ssim(img1, img2):
    C1 = (0.01 * 255)**2
    C2 = (0.03 * 255)**2
    # C1 = (0.01)**2
    # C2 = (0.03)**2
    img1 = img1.astype(np.float64)
    img2 = img2.astype(np.float64)
    kernel = cv2.getGaussianKernel(11, 1.5)
    window = np.outer(kernel, kernel.transpose())

    mu1 = cv2.filter2D(img1, -1, window)[5:-5, 5:-5]  # valid
    mu2 = cv2.filter2D(img2, -1, window)[5:-5, 5:-5]
    mu1_sq = mu1**2
    mu2_sq = mu2**2
    mu1_mu2 = mu1 * mu2
    sigma1_sq = cv2.filter2D(img1**2, -1, window)[5:-5, 5:-5] - mu1_sq
    sigma2_sq = cv2.filter2D(img2**2, -1, window)[5:-5, 5:-5] - mu2_sq
    sigma12 = cv2.filter2D(img1 * img2, -1, window)[5:-5, 5:-5] - mu1_mu2

    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) *
                                                            (sigma1_sq + sigma2_sq + C2))
    return ssim_map.mean()


def calculate_ssim(img1, img2):
    '''calculate SSIM
    the same outputs as MATLAB's
    img1, img2: [0, 255]
    '''
    if not img1.shape == img2.shape:
        raise ValueError('Input images must have the same dimensions.')
    if img1.ndim == 2:
        return ssim(img1, img2)
    elif img1.ndim == 3:
        if img1.shape[2] == 3:
            ssims = []
            for i in range(3):
                ssims.append(ssim(img1, img2))
            return np.array(ssims).mean()
        elif img1.shape[2] == 1:
            return ssim(np.squeeze(img1), np.squeeze(img2))
    else:
        raise ValueError('Wrong input image dimensions.')

"""
Creates 'pdist' model, which we need to calcuate 'pdist' metric during validation
"""
def create_pdist_model(network='vgg', use_gpu=True):
    '''
    percetual similarity metric
    https://github.com/richzhang/PerceptualSimilarity
    :param network can be vgg or alex
    '''
    print("Initializing pdist model")
    print("-"*40)
    # Initializing the model
    model = dm.DistModel()
    model.initialize(model='net-lin', net=network, use_gpu=use_gpu)
    return model


def calculate_pdist(model, img1, img2):
    # expand channel to 3 dim for vgg
    img1 = np.expand_dims(img1, axis=2)
    img1 = np.repeat(img1, 3, axis=2)
    img2 = np.expand_dims(img2, axis=2)
    img2 = np.repeat(img2, 3, axis=2)
    #  image from [-1,1]
    img1 = img2tensor(img1)
    img2 = img2tensor(img2)

    # Compute distance
    with torch.no_grad():
        dist = model.forward(img1, img2)
    return dist[0]

def cvt_int(self, inputString):
    find_digit = re.search(r'\d', inputString)
    reformat_num = int(float(inputString[find_digit.start():]))
    reformat_str = inputString[:find_digit.start()] + str(reformat_num)
    return reformat_str

def load_dicom(folder_location):
    '''
    :function 
    Load dicom images in one folder, give them correct ordering. 
    :returns
    A 3D volume and pixel spacings 
    '''
    # retrieve names of all dcm files
    dicom_names = glob.glob(os.path.join(folder_location,'*.dcm'))
    spacings = [] 
    # load dcm files 
    files = []
    for fname in dicom_names:
        files.append(pydicom.dcmread(fname))

    # skip files with no SliceLocation 
    slices = []
    skipcount = 0
    for f in files:
        if hasattr(f, 'SliceLocation'):
            slices.append(f)
        else:
            skipcount = skipcount + 1

    # print("skipped, no SliceLocation: {}".format(skipcount))

    # ensure slices are in the correct order (sorted)
    slices = sorted(slices, key=lambda s: s.SliceLocation)

    # pixel spacings, assuming all slices are the same
    ps = [0, 0, 0]
    ps[0], ps[1] = float(slices[0].PixelSpacing[0]), float(slices[0].PixelSpacing[1])
    cosines0 = slices[0].ImageOrientationPatient
    iip0 = slices[0].ImagePositionPatient
    normal0 = [cosines0[1]*cosines0[5] - cosines0[2]*cosines0[4], 
                cosines0[2]*cosines0[3] - cosines0[0]*cosines0[5],
                cosines0[0]*cosines0[4] - cosines0[1]*cosines0[3]]
    dist0 = 0 
    for i in range(3): 
        dist0 += normal0[i]*iip0[i]
    
    iip1 = slices[1].ImagePositionPatient
    dist1 = 0 
    for i in range(3): 
        dist1 += normal0[i]*iip1[i]

    # subtract and take abs to obtain pixel spacing in z direction 
    ps[2] = abs(dist1 - dist0)

    # create 3D volume in [D,H,W]
    # img_shape = list(slices[0].pixel_array.shape)
    # img_shape.append(len(slices))
    img_shape = [len(slices), slices[0].pixel_array.shape[0], slices[0].pixel_array.shape[1]]
    vol = np.zeros(img_shape)

    # fill 3D array with the images from the files
    for i, s in enumerate(slices):
        img2d = s.pixel_array
        vol[i,:,:] = img2d
        
    return vol, ps 


def get_dcm_paths(data_root): 
        paths = [] 
        ids = [] 
        for patient in os.listdir(data_root): 
            patient_path = os.path.join(data_root, patient)
            if not os.path.isdir(patient_path): 
                continue 
            for timepoint in os.listdir(patient_path): 
                time_path = os.path.join(patient_path, timepoint)
                for f in os.listdir(time_path): 
                    paths.append( os.path.join(time_path, f, 'DICOM') )
                    ids.append(f.split('-')[1]) # [-1]
        return paths, ids


import os, os.path
import glob
import math
import argparse
import random
import logging
import torch
import options.options as option
from utils import util
from data import create_dataloader, create_dataset
from models import create_model
import numpy as np
from collections import OrderedDict


def main():
    # argument parser
    parser = argparse.ArgumentParser()
    parser.add_argument('-opt', type=str, required=True, help='Path to option JSON file.')
    opt = option.parse(parser.parse_args().opt, is_train=True)
    # convert to a dictionary with 'None' for missing key
    opt = option.dict_to_nonedict(opt)
    torch.backends.cudnn.benchmark = True
 
    if opt['path']['resume_state']:
        resume_state = torch.load(opt['path']['resume_state'])
    else:
        resume_state = None
        util.mkdir_and_rename(opt['path']['experiments_root'])
        util.mkdirs((path for key, path in opt['path'].items() if not key == 'experiments_root'
                     and 'pretrain_model' not in key and 'resume' not in key))

    # config loggers. without it, logger will not work
    util.setup_logger('base', opt['path']['log'], 'train', level=logging.INFO, screen=True)
    util.setup_logger('val', opt['path']['log'], 'val', level=logging.INFO)
    logger = logging.getLogger('base') # get the 'base' logger which records training data
    
    # if 'resume_state' is not None, start training from the end state of weights
    if resume_state:
        logger.info('Resuming training from epoch: {}, iter: {}.'.format(
            resume_state['epoch'], resume_state['iter']))
        # check resume options and paths for model weights are available
        option.check_resume(opt)  
    
    # print the parsed dictionary into readable form and print on screen
    logger.info(option.dict2str(opt))

    # set up tensorboard logger if path specified and 'debug' not in name
    if opt['use_tb_logger'] and 'debug' not in opt['name']:
        from torch.utils.tensorboard import SummaryWriter
        # create summary writer
        tb_logger = SummaryWriter(log_dir='../tb_logger/' + opt['name'])
    seed = opt['train']['manual_seed']
    if seed is None:
        seed = random.randint(1, 10000)
    logger.info('Random seed: {}'.format(seed))
    util.set_random_seed(seed)

    need_label, data_merged = opt['need_label'], opt['data_merged'] 
    need_atten = opt['network_G']['use_attention']
    if need_atten:
        need_candidate_atten = opt['network_G']['use_attention']['generate_candidate']
    else:
        need_candidate_atten = opt['network_G']['use_attention']
        
    # 'train' and 'val' keys in opt['datasets'] with parameters for their respective phase
    for phase, dataset_opt in opt['datasets'].items():
        dataset_opt['need_label'] = need_label
        dataset_opt['data_merged'] = data_merged
        dataset_opt['use_candidate_atten'] = need_candidate_atten
        if phase == 'train':
            train_set = create_dataset(dataset_opt)
            train_size = int(math.floor(len(train_set) / dataset_opt['batch_size']))
            logger.info('Number of train cases: {:,d}, iters required: total cases/batch size = {}/{}={}'.format(
                len(train_set), len(train_set), dataset_opt['batch_size'], train_size))
            total_iters = int(opt['train']['niter'])
            print("Total training iterations:", total_iters)
            total_epochs = int(math.ceil(total_iters / train_size))
            logger.info('Total epochs needed: {:,d}/{:,d} = {:,d} for iters {:,d}'.format(
                total_iters, train_size, total_epochs, total_iters))
            train_loader = create_dataloader(train_set, dataset_opt)
        elif phase == 'val':
            val_set = create_dataset(dataset_opt)
            val_loader = create_dataloader(val_set, dataset_opt)
            logger.info('Number of val cases in [{:s}]: {:d}'.format(dataset_opt['name'],
                                                                      len(val_set)))
        else:
            raise NotImplementedError('Phase [{:s}] is not recognized.'.format(phase))
    assert train_loader is not None

    """
    data = iter(train_loader).next()
    print("batch data keys:", data.keys())
    print('LR shape:', data['LR'].shape)
    print('HR shape:', data['HR'].shape)
    print('Mask shape:', data['mask'].shape)
    print('Candidates shape:', data['Cand'].shape)
    print('uid:', data['uid'])
    print('kernel:', data['kernel'])
    print('dose:', data['dose'])
    print('-'*40)
    data = iter(val_loader).next()
    print("batch data keys:", data.keys())
    print('LR shape:', data['LR'].shape)
    print('HR shape:', data['HR'].shape)
    print('Mask shape:', data['mask'].shape)
    print('Candidates shape:', data['Cand'].shape)
    """

    # create model (by default, model is already fp32)
    model = create_model(opt)
    # initialize amp for mixed precision training, determines best precision for each layer
    model.initialize_amp()

    if resume_state:
        start_epoch = resume_state['epoch'] # get last epoch training was stopped on, which becomes start epoch
        current_step = resume_state['iter'] # get number of iterations from last training
        model.resume_training(resume_state)  # handle optimizers and schedulers
    else:
        current_step = 0 
        start_epoch = 0

    print('Model Instantiated!')
    ## =====================
    ## START TRAINING
    ## =====================
    logger.info('Start training from epoch: {:d}, iter: {:d}'.format(start_epoch, current_step))
    for epoch in range(start_epoch, total_epochs):
        # print("Epoch:", epoch)
        for _, train_data in enumerate(train_loader):
            current_step += 1
            if current_step > total_iters:
                break
            # set up attributes: 'self.var_L' & 'self.real_H' for 'LR' &  'HR' image respectively
            model.feed_train_data(train_data)
            model.optimize_parameters(current_step)
            # update learning rate, called after .step() after pytorch 1.1
            model.update_learning_rate()
            # print("check for logging: {} % {} == 0 = {}".format(current_step, opt['logger']['print_freq'],
            # current_step % opt['logger']['print_freq'] == 0))
            # log : opt['logger']['print_freq'] == 10, so logger will print every 10 step
            
            if current_step % opt['logger']['print_freq'] == 0:
                logs = model.get_current_log() # gets dictionary that contains gan losses
                # 'get_current_learning_rate()' returns current lr for generator
                message = '<epoch:{:5d}, iter:{:8,d}, lr:{:.6e}> '.format(
                    epoch, current_step, model.get_current_learning_rate())
                
                # 'logs.items()' give (key, value) pairs, we add that to the initial message we set up above
                for k, v in logs.items():
                    message += '{:s}: {:.4e} '.format(k, v)
                    # add log items (losses) to tensorboard logger and print message on screen
                    if opt['use_tb_logger'] and 'debug' not in opt['name']:
                        tb_logger.add_scalar(k, v, current_step)
                logger.info(message)

            # save models and training states at every that step
            if current_step % opt['logger']['save_checkpoint_freq'] == 0:
                logger.info('Saving models and training states.')
                model.save(current_step) # save 'state_dict' of both G & D at given step
                # save 'epoch', 'current_step', 'optimizer', 'scheduler'
                model.save_training_state(epoch, current_step) 

            # validation loop
            # print("checking to run validation: {} % {} = {}".format(current_step, opt['train']['val_freq'],
            if current_step % opt['train']['val_freq'] == 0:
                # copy net and convert to fp16 if necessary, according to json
                model.half()
                # create pdist model (PNetLin - based on vgg16)
                pdist_model = util.create_pdist_model()
                # 'OrderedDictionary' to store evaluation metrics
                pnsr_results = OrderedDict()
                ssim_results = OrderedDict()
                pdist_results = OrderedDict()
                for val_data in val_loader:
                    logger.info('start inference...')
                    model.feed_test_data(val_data)
                    model.test(val_data)  # test
                    # get cpu numpy from cuda tensor
                    has_mask = False if val_loader.dataset.opt['maskroot_HR'] is None else True
                    visuals = model.get_current_visuals(val_data, maskOn=has_mask)
                    # save volume data
                    patient_id = val_data['uid'][0]
                    util.mkdir(os.path.join(opt['path']['val_images'], str(current_step)))
                    vol_path = os.path.join(opt['path']['val_images'], str(current_step), patient_id)
                    LR_spacings = [x.item() for x in val_data['spacings']] if val_data['spacings'] else None
                    if opt['result_format'] == 'nrrd':
                        logger.info('saving nnrd...')
                        sr_vol = util.tensor2img(visuals['SR'], out_type=np.uint16) 
                        util.save_vol(opt, LR_spacings, vol_path + '.nrrd', sr_vol)
                    elif opt['result_format'] == 'dicom':
                        logger.info('saving dicoms...') 
                        sr_vol = util.tensor2img(visuals['SR'], out_type=np.int16, intercept = -1000) 
                        util.save_dicoms(opt, LR_spacings, vol_path, sr_vol)
                    else: 
                        raise NotImplementedError('supported output format: nrrd or dicom')

                    # initialize dictionary with empty {}
                    pnsr_results[patient_id] = {}
                    ssim_results[patient_id] = {}
                    pdist_results[patient_id] = {}

                    def _calculate_metrics(sr_vol, gt_vol, view='xy'):
                        sum_psnr = 0.
                        sum_ssim = 0.
                        sum_pdist = 0.
                        # [D,H,W]
                        num_val = 0  # psnr could be inf at xz or yz (near edges), will not calculate
                        for i, vol in enumerate(zip(sr_vol, gt_vol)):
                            sr_img, gt_img = vol[0], vol[1]
                            # calculate PSNR and SSIM
                            # range is assume to be [0,255] so  have to scale back from 1500 to 255 float64
                            crop_size = round(opt['scale'])
                            cropped_sr_img = sr_img[crop_size:-crop_size, crop_size:-crop_size] \
                                                 .astype(np.float64) / 1500. * 255.
                            cropped_gt_img = gt_img[crop_size:-crop_size, crop_size:-crop_size] \
                                                 .astype(np.float64) / 1500. * 255.

                            psnr = util.calculate_psnr(cropped_sr_img, cropped_gt_img)
                            ssim = util.calculate_ssim(cropped_sr_img, cropped_gt_img)
                            pdist = util.calculate_pdist(pdist_model, cropped_sr_img, cropped_gt_img)

                            if psnr != float('inf'):
                                num_val += 1
                                sum_psnr += psnr
                                sum_ssim += ssim
                                sum_pdist += pdist
                            logger.info('{:20s} - {:3d}- PSNR: {:.6f} dB; SSIM: {:.6f}; pdist: {:.6f}.' \
                                        .format(patient_id, i + 1, psnr, ssim, pdist))

                        pnsr_results[patient_id][view] = sum_psnr / num_val
                        ssim_results[patient_id][view] = sum_ssim / num_val
                        pdist_results[patient_id][view] = sum_pdist / num_val
                        return pnsr_results, ssim_results, pdist_results

                    sr_vol = util.tensor2img(visuals['SR'], out_type=np.uint16) 
                    gt_vol = util.tensor2img(visuals['HR'], out_type=np.uint16) # uint16 range [0,1500]
                    # [H W] axial view
                    _calculate_metrics(sr_vol, gt_vol, view='xy')
                    # [D W] coronal view
                    _calculate_metrics(sr_vol.transpose(1, 0, 2), gt_vol.transpose(1, 0, 2), view='xz')
                    # [D H] sagittal view
                    _calculate_metrics(sr_vol.transpose(2, 0, 1), gt_vol.transpose(2, 0, 1), view='yz')

                logger_val = logging.getLogger('val')  # validation logger
                logger_val.info('<epoch:{:5d}, iter:{:8,d} > '.format(epoch, current_step))
                # prints the metrics to screen
                avg_psnr = util.print_metrics(logger_val, 'val PSNR', pnsr_results)
                avg_ssim = util.print_metrics(logger_val, 'val SSIM', ssim_results)
                avg_pdist = util.print_metrics(logger_val, 'val pdist', pdist_results)

                # record evluation metrics on tensorboard
                if opt['use_tb_logger'] and 'debug' not in opt['name']:
                    tb_logger.add_scalars('psnr', avg_psnr, current_step)
                    tb_logger.add_scalars('ssim', avg_ssim, current_step)
                    tb_logger.add_scalars('pdist', avg_pdist, current_step)

    logger.info('Saving the final model.')
    model.save('latest')
    logger.info('End of training.')


if __name__ == '__main__':
    main()



import sys
from shutil import get_terminal_size
import time


class ProgressBar(object):
    '''A progress bar which can print the progress
    modified from https://github.com/hellock/cvbase/blob/master/cvbase/progress.py
    '''

    def __init__(self, task_num=0, bar_width=50, start=True):
        self.task_num = task_num
        max_bar_width = self._get_max_bar_width()
        self.bar_width = (bar_width if bar_width <= max_bar_width else max_bar_width)
        self.completed = 0
        if start:
            self.start()

    def _get_max_bar_width(self):
        terminal_width, _ = get_terminal_size()
        max_bar_width = min(int(terminal_width * 0.6), terminal_width - 50)
        if max_bar_width < 10:
            print('terminal width is too small ({}), please consider widen the terminal for better '
                  'progressbar visualization'.format(terminal_width))
            max_bar_width = 10
        return max_bar_width

    def start(self):
        if self.task_num > 0:
            sys.stdout.write('[{}] 0/{}, elapsed: 0s, ETA:\n{}\n'.format(
                ' ' * self.bar_width, self.task_num, 'Start...'))
        else:
            sys.stdout.write('completed: 0, elapsed: 0s')
        sys.stdout.flush()
        self.start_time = time.time()

    def update(self, msg='In progress...'):
        self.completed += 1
        elapsed = time.time() - self.start_time
        fps = self.completed / elapsed
        if self.task_num > 0:
            percentage = self.completed / float(self.task_num)
            eta = int(elapsed * (1 - percentage) / percentage + 0.5)
            mark_width = int(self.bar_width * percentage)
            bar_chars = '█' * mark_width + '-' * (self.bar_width - mark_width)
            sys.stdout.write('\033[2F')  # cursor up 2 lines
            sys.stdout.write('\033[J')  # clean the output (remove extra chars since last display)
            sys.stdout.write('[{}] {}/{}, {:.1f} task/s, elapsed: {}s, ETA: {:5}s\n{}\n'.format(
                bar_chars, self.completed, self.task_num, fps, int(elapsed + 0.5), eta, msg))
        else:
            sys.stdout.write('completed: {}, elapsed: {}s, {:.1f} tasks/s'.format(
                self.completed, int(elapsed + 0.5), fps))
        sys.stdout.flush()
